from unittest import TestCase, mock

from ......protocols.didcomm_prefix import DIDCommPrefix
from ...message_types import PROBLEM_REPORT, PROTOCOL_PACKAGE
from ...messages.problem_report import RotateProblemReport


class TestRotateProblemReport(TestCase):
    def test_init_type(self):
        """Test initializer."""

        obj = RotateProblemReport()
        assert obj._type == DIDCommPrefix.qualify_current(PROBLEM_REPORT)

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.problem_report.RotateProblemReportSchema.load"
    )
    def test_deserialize(self, mock_rotate_ack_schema_load):
        """Test deserialization."""

        obj = RotateProblemReport()
        rotate_ack = RotateProblemReport.deserialize(obj)
        mock_rotate_ack_schema_load.assert_called_once_with(obj)
        assert rotate_ack is mock_rotate_ack_schema_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.problem_report.RotateProblemReportSchema.dump"
    )
    def test_serialize(self, mock_rotate_ack_schema_dump):
        """Test serialization."""

        obj = RotateProblemReport()
        rotate_ack_dict = obj.serialize()
        mock_rotate_ack_schema_dump.assert_called_once_with(obj)
        assert rotate_ack_dict is mock_rotate_ack_schema_dump.return_value

    def test_serde(self):
        obj = {
            "~thread": {"thid": "test-thid", "pthid": "test-pthid"},
            "problem_items": [{"did": "test-did", "reason": "test-reason"}],
        }
        rotate_ack = RotateProblemReport.deserialize(obj)
        assert rotate_ack._type == DIDCommPrefix.qualify_current(PROBLEM_REPORT)

        rotate_ack_dict = rotate_ack.serialize()
        assert rotate_ack_dict["problem_items"] == obj["problem_items"]

    def test_make_model(self):
        """Test making model."""

        obj = RotateProblemReport(
            problem_items=[{"did": "test-did", "reason": "test-reason"}]
        )
        obj.assign_thread_id("test-thid", "test-pthid")
        rotate_ack_dict = obj.serialize()
        rotate_ack = RotateProblemReport.deserialize(rotate_ack_dict)
        assert isinstance(rotate_ack, RotateProblemReport)



"""Sign and verify functions for json-ld based credentials."""

import json

from ...core.profile import ProfileSession
from ...did.did_key import DIDKey
from ...vc.ld_proofs import DocumentLoader
from ...wallet.base import BaseWallet
from ...wallet.key_type import ED25519
from ...wallet.util import b64_to_bytes, b64_to_str, bytes_to_b64, str_to_b64
from .create_verify_data import create_verify_data
from .error import BadJWSHeaderError


def did_key(verkey: str) -> str:
    """Qualify verkey into DID key if need be."""

    if verkey.startswith("did:key:"):
        return verkey

    return DIDKey.from_public_key_b58(verkey, ED25519).did


def b64encode(val: str) -> str:
    """Url Safe B64 Encode."""
    return str_to_b64(val, urlsafe=True, pad=False)


def b64decode(val: str) -> str:
    """Url Safe B64 Decode."""
    return b64_to_str(val, urlsafe=True)


def create_jws(encoded_header: str, verify_data: bytes) -> bytes:
    """Compose JWS."""
    return (encoded_header + ".").encode("utf-8") + verify_data


async def jws_sign(session: ProfileSession, verify_data: bytes, verkey: str) -> str:
    """Sign JWS."""

    header = {"alg": "EdDSA", "b64": False, "crit": ["b64"]}

    encoded_header = b64encode(json.dumps(header))

    jws_to_sign = create_jws(encoded_header, verify_data)

    wallet = session.inject(BaseWallet)
    signature = await wallet.sign_message(jws_to_sign, verkey)

    encoded_signature = bytes_to_b64(signature, urlsafe=True, pad=False)

    return encoded_header + ".." + encoded_signature


def verify_jws_header(header: dict) -> None:
    """Check header requirements."""

    if header != {"alg": "EdDSA", "b64": False, "crit": ["b64"]}:
        raise BadJWSHeaderError("Invalid JWS header parameters for Ed25519Signature2018.")


async def jws_verify(
    session: ProfileSession, verify_data: bytes, signature: str, public_key: str
) -> bool:
    """Detached jws verify handling."""

    encoded_header, _, encoded_signature = signature.partition("..")
    decoded_header = json.loads(b64decode(encoded_header))

    verify_jws_header(decoded_header)

    decoded_signature = b64_to_bytes(encoded_signature, urlsafe=True)

    jws_to_verify = create_jws(encoded_header, verify_data)

    wallet = session.inject(BaseWallet)
    verified = await wallet.verify_message(
        jws_to_verify, decoded_signature, public_key, ED25519
    )

    return verified


async def sign_credential(
    session: ProfileSession, credential: dict, signature_options: dict, verkey: str
) -> dict:
    """Sign Credential."""

    document_loader = session.profile.inject_or(DocumentLoader)
    _, verify_data_hex_string = create_verify_data(
        credential,
        signature_options,
        document_loader,
    )
    verify_data_bytes = bytes.fromhex(verify_data_hex_string)
    jws = await jws_sign(session, verify_data_bytes, verkey)
    return {**credential, "proof": {**signature_options, "jws": jws}}


async def verify_credential(session: ProfileSession, doc: dict, verkey: str) -> bool:
    """Verify credential."""

    document_loader = session.profile.inject_or(DocumentLoader)
    framed, verify_data_hex_string = create_verify_data(
        doc,
        doc["proof"],
        document_loader,
    )
    verify_data_bytes = bytes.fromhex(verify_data_hex_string)
    valid = await jws_verify(session, verify_data_bytes, framed["proof"]["jws"], verkey)
    return valid



from ...ld_proofs import DocumentVerificationResult, ProofResult, PurposeResult

# All signed documents manually tested for validity on https://univerifier.io

CREDENTIAL_TEMPLATE = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
    ],
    "id": "http://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:456",
        "degree": {"type": "BachelorDegree", "name": "Bachelor of Science and Arts"},
    },
}

CREDENTIAL_ISSUED = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
    ],
    "id": "http://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:456",
        "degree": {"type": "BachelorDegree", "name": "Bachelor of Science and Arts"},
    },
    "proof": {
        "type": "Ed25519Signature2018",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2019-12-11T03:50:55+00:00",
        "proofPurpose": "assertionMethod",
        "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..tELgWSCo0jSOKLpHWLOe2ZWj6WxLFc6FcqBZzgQFdcaawC5mARYmdr0XA37TMUp2Q9jUwriNP4gQm3GFIHV9BQ",
    },
}

CREDENTIAL_TEMPLATE_2020 = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/suites/ed25519-2020/v1",
    ],
    "id": "http://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:456",
        "degree": {"type": "BachelorDegree", "name": "Bachelor of Science and Arts"},
    },
}

CREDENTIAL_ISSUED_2020 = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/suites/ed25519-2020/v1",
    ],
    "id": "http://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:456",
        "degree": {"type": "BachelorDegree", "name": "Bachelor of Science and Arts"},
    },
    "proof": {
        "type": "Ed25519Signature2020",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2019-12-11T03:50:55+00:00",
        "proofPurpose": "assertionMethod",
        "proofValue": "z3AkzFbAERyy7KwTDbafaMjoeYejfnzMfjV1eG93j1yhsPGGTJWeyxcWVexxg84AsnRDmMFeVN6SN2dbxCsTsn2eY",
    },
}

CREDENTIAL_TEMPLATE_BBS = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://w3id.org/security/bbs/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
    ],
    "id": "http://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": {
        "id": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa"
    },
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:456",
        "degree": {"type": "BachelorDegree", "name": "Bachelor of Science and Arts"},
    },
}

CREDENTIAL_ISSUED_BBS = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://w3id.org/security/bbs/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
    ],
    "id": "http://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": {
        "id": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa"
    },
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:456",
        "degree": {"type": "BachelorDegree", "name": "Bachelor of Science and Arts"},
    },
    "proof": {
        "type": "BbsBlsSignature2020",
        "verificationMethod": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
        "created": "2019-12-11T03:50:55+00:00",
        "proofPurpose": "assertionMethod",
        "proofValue": "ke/4UsKMbYyFQP2WAsQs0QGJ0PCfvN3tWMoD8oBjroY3mrgXA7FmUqVy0HaZIoKvFHl1FnE1jqh0OnqErooFBMTu0a4qe3klziRSuRuLdUhgbgzOJeJiWPlrs6Y5/D5d0CRp8ifPj7xZBIy3Iuoz4g==",
    },
}

CREDENTIAL_VERIFIED = DocumentVerificationResult(
    verified=True,
    document={
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "id": "http://example.gov/credentials/3732",
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
        "issuanceDate": "2020-03-10T04:24:12.164Z",
        "credentialSubject": {
            "id": "did:example:456",
            "degree": {
                "type": "BachelorDegree",
                "name": "Bachelor of Science and Arts",
            },
        },
        "proof": {
            "type": "Ed25519Signature2018",
            "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "created": "2019-12-11T03:50:55+00:00",
            "proofPurpose": "assertionMethod",
            "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..tELgWSCo0jSOKLpHWLOe2ZWj6WxLFc6FcqBZzgQFdcaawC5mARYmdr0XA37TMUp2Q9jUwriNP4gQm3GFIHV9BQ",
        },
    },
    results=[
        ProofResult(
            verified=True,
            proof={
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2019-12-11T03:50:55+00:00",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..tELgWSCo0jSOKLpHWLOe2ZWj6WxLFc6FcqBZzgQFdcaawC5mARYmdr0XA37TMUp2Q9jUwriNP4gQm3GFIHV9BQ",
            },
            purpose_result=PurposeResult(
                valid=True,
                controller={
                    "@context": "https://w3id.org/security/v2",
                    "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "assertionMethod": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "authentication": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                        }
                    ],
                    "capabilityDelegation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "capabilityInvocation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "keyAgreement": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
                            "type": "X25519KeyAgreementKey2019",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
                        }
                    ],
                    "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                },
            ),
        )
    ],
)

CREDENTIAL_VERIFIED_2020 = DocumentVerificationResult(
    verified=True,
    document={
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
            "https://w3id.org/security/suites/ed25519-2020/v1",
        ],
        "id": "http://example.gov/credentials/3732",
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
        "issuanceDate": "2020-03-10T04:24:12.164Z",
        "credentialSubject": {
            "id": "did:example:456",
            "degree": {
                "type": "BachelorDegree",
                "name": "Bachelor of Science and Arts",
            },
        },
        "proof": {
            "type": "Ed25519Signature2020",
            "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "created": "2019-12-11T03:50:55+00:00",
            "proofPurpose": "assertionMethod",
            "proofValue": "z3AkzFbAERyy7KwTDbafaMjoeYejfnzMfjV1eG93j1yhsPGGTJWeyxcWVexxg84AsnRDmMFeVN6SN2dbxCsTsn2eY",
        },
    },
    results=[
        ProofResult(
            verified=True,
            proof={
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                    "https://w3id.org/security/suites/ed25519-2020/v1",
                ],
                "type": "Ed25519Signature2020",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2019-12-11T03:50:55+00:00",
                "proofPurpose": "assertionMethod",
                "proofValue": "z3AkzFbAERyy7KwTDbafaMjoeYejfnzMfjV1eG93j1yhsPGGTJWeyxcWVexxg84AsnRDmMFeVN6SN2dbxCsTsn2eY",
            },
            purpose_result=PurposeResult(
                valid=True,
                controller={
                    "@context": "https://w3id.org/security/v2",
                    "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "assertionMethod": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "authentication": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                        }
                    ],
                    "capabilityDelegation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "capabilityInvocation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "keyAgreement": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
                            "type": "X25519KeyAgreementKey2019",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
                        }
                    ],
                    "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                },
            ),
        )
    ],
)

PRESENTATION_UNSIGNED = {
    "@context": ["https://www.w3.org/2018/credentials/v1"],
    "type": ["VerifiablePresentation"],
    "verifiableCredential": [
        {
            "@context": [
                "https://www.w3.org/2018/credentials/v1",
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "id": "http://example.gov/credentials/3732",
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
            "issuanceDate": "2020-03-10T04:24:12.164Z",
            "credentialSubject": {
                "id": "did:example:456",
                "degree": {
                    "type": "BachelorDegree",
                    "name": "Bachelor of Science and Arts",
                },
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2019-12-11T03:50:55+00:00",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..tELgWSCo0jSOKLpHWLOe2ZWj6WxLFc6FcqBZzgQFdcaawC5mARYmdr0XA37TMUp2Q9jUwriNP4gQm3GFIHV9BQ",
            },
        }
    ],
}

PRESENTATION_SIGNED = {
    "@context": ["https://www.w3.org/2018/credentials/v1"],
    "type": ["VerifiablePresentation"],
    "verifiableCredential": [
        {
            "@context": [
                "https://www.w3.org/2018/credentials/v1",
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "id": "http://example.gov/credentials/3732",
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
            "issuanceDate": "2020-03-10T04:24:12.164Z",
            "credentialSubject": {
                "id": "did:example:456",
                "degree": {
                    "type": "BachelorDegree",
                    "name": "Bachelor of Science and Arts",
                },
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2019-12-11T03:50:55+00:00",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..tELgWSCo0jSOKLpHWLOe2ZWj6WxLFc6FcqBZzgQFdcaawC5mARYmdr0XA37TMUp2Q9jUwriNP4gQm3GFIHV9BQ",
            },
        }
    ],
    "proof": {
        "type": "Ed25519Signature2018",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2020-12-11T03:50:55+00:00",
        "proofPurpose": "authentication",
        "challenge": "2b1bbff6-e608-4368-bf84-67471b27e41c",
        "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..vLk6iXrPxdt4xWq_Lwnd85cy17Npol1ALxYoMhZrKpmkaWDOoz_exP1ggurPik7Rhpa7a82AfrT0OnigJnJsAQ",
    },
}



"""Revocation registry admin routes."""

import json
import logging

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate, validates_schema
from marshmallow.exceptions import ValidationError
from uuid_utils import uuid4

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..anoncreds.base import (
    AnonCredsObjectNotFound,
    AnonCredsRegistrationError,
    AnonCredsResolutionError,
)
from ..anoncreds.default.legacy_indy.registry import LegacyIndyRegistry
from ..anoncreds.issuer import AnonCredsIssuerError
from ..anoncreds.models.revocation import RevRegDefState
from ..anoncreds.revocation import AnonCredsRevocation, AnonCredsRevocationError
from ..anoncreds.routes import (
    AnonCredsRevocationModuleResponseSchema,
    AnonCredsRevRegIdMatchInfoSchema,
    create_transaction_for_endorser_description,
    endorser_connection_id_description,
)
from ..askar.profile_anon import AskarAnonCredsProfile
from ..indy.issuer import IndyIssuerError
from ..indy.models.revocation import IndyRevRegDef
from ..ledger.base import BaseLedger
from ..ledger.error import LedgerError
from ..ledger.multiple_ledger.base_manager import BaseMultipleLedgerManager
from ..messaging.models.openapi import OpenAPISchema
from ..messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_CRED_DEF_ID_VALIDATE,
    ANONCREDS_CRED_REV_ID_EXAMPLE,
    ANONCREDS_CRED_REV_ID_VALIDATE,
    ANONCREDS_REV_REG_ID_EXAMPLE,
    ANONCREDS_REV_REG_ID_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
    WHOLE_NUM_EXAMPLE,
    WHOLE_NUM_VALIDATE,
    UUIDFour,
)
from ..revocation.error import RevocationError
from ..revocation.models.issuer_rev_reg_record import (
    IssuerRevRegRecord,
    IssuerRevRegRecordSchema,
)
from ..storage.error import StorageError, StorageNotFoundError
from ..utils.profiles import is_not_anoncreds_profile_raise_web_exception
from .manager import RevocationManager, RevocationManagerError
from .models.issuer_cred_rev_record import (
    IssuerCredRevRecord,
    IssuerCredRevRecordSchemaAnonCreds,
)

LOGGER = logging.getLogger(__name__)

TAG_TITLE = "AnonCreds - Revocation"


class RevRegResultSchemaAnonCreds(OpenAPISchema):
    """Result schema for revocation registry creation request."""

    result = fields.Nested(IssuerRevRegRecordSchema())


class CredRevRecordQueryStringSchema(OpenAPISchema):
    """Parameters and validators for credential revocation record request."""

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields - must have (rr-id and cr-id) xor cx-id."""

        rev_reg_id = data.get("rev_reg_id")
        cred_rev_id = data.get("cred_rev_id")
        cred_ex_id = data.get("cred_ex_id")

        if not (
            (rev_reg_id and cred_rev_id and not cred_ex_id)
            or (cred_ex_id and not rev_reg_id and not cred_rev_id)
        ):
            raise ValidationError(
                "Request must have either rev_reg_id and cred_rev_id or cred_ex_id"
            )

    rev_reg_id = fields.Str(
        required=False,
        validate=ANONCREDS_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": ANONCREDS_REV_REG_ID_EXAMPLE,
        },
    )
    cred_rev_id = fields.Str(
        required=False,
        validate=ANONCREDS_CRED_REV_ID_VALIDATE,
        metadata={
            "description": "Credential revocation identifier",
            "example": ANONCREDS_CRED_REV_ID_EXAMPLE,
        },
    )
    cred_ex_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Credential exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )


class RevRegId(OpenAPISchema):
    """Parameters and validators for delete tails file request."""

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields - must have either rr-id or cr-id."""

        rev_reg_id = data.get("rev_reg_id")
        cred_def_id = data.get("cred_def_id")

        if not (rev_reg_id or cred_def_id):
            raise ValidationError("Request must have either rev_reg_id or cred_def_id")

    rev_reg_id = fields.Str(
        required=False,
        validate=ANONCREDS_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": ANONCREDS_REV_REG_ID_EXAMPLE,
        },
    )
    cred_def_id = fields.Str(
        required=False,
        validate=ANONCREDS_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )


class CredRevRecordResultSchemaAnonCreds(OpenAPISchema):
    """Result schema for credential revocation record request."""

    result = fields.Nested(IssuerCredRevRecordSchemaAnonCreds())


class CredRevRecordDetailsResultSchemaAnonCreds(OpenAPISchema):
    """Result schema for credential revocation record request."""

    results = fields.List(fields.Nested(IssuerCredRevRecordSchemaAnonCreds()))


class CredRevIndyRecordsResultSchemaAnonCreds(OpenAPISchema):
    """Result schema for revoc reg delta."""

    rev_reg_delta = fields.Dict(
        metadata={"description": "Indy revocation registry delta"}
    )


class RevRegIssuedResultSchemaAnonCreds(OpenAPISchema):
    """Result schema for revocation registry credentials issued request."""

    result = fields.Int(
        validate=WHOLE_NUM_VALIDATE,
        metadata={
            "description": "Number of credentials issued against revocation registry",
            "strict": True,
            "example": WHOLE_NUM_EXAMPLE,
        },
    )


class RevRegUpdateRequestMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking rev reg id."""

    apply_ledger_update = fields.Bool(
        required=True,
        metadata={"description": "Apply updated accumulator transaction to ledger"},
    )


class RevRegWalletUpdatedResultSchemaAnonCreds(OpenAPISchema):
    """Number of wallet revocation entries status updated."""

    rev_reg_delta = fields.Dict(
        metadata={"description": "Indy revocation registry delta"}
    )
    accum_calculated = fields.Dict(
        metadata={"description": "Calculated accumulator for phantom revocations"}
    )
    accum_fixed = fields.Dict(
        metadata={"description": "Applied ledger transaction to fix revocations"}
    )


class RevRegsCreatedSchemaAnonCreds(OpenAPISchema):
    """Result schema for request for revocation registries created."""

    rev_reg_ids = fields.List(
        fields.Str(
            validate=ANONCREDS_REV_REG_ID_VALIDATE,
            metadata={
                "description": "Revocation registry identifiers",
                "example": ANONCREDS_REV_REG_ID_EXAMPLE,
            },
        )
    )


class RevRegUpdateTailsFileUriSchema(OpenAPISchema):
    """Request schema for updating tails file URI."""

    tails_public_uri = fields.Url(
        required=True,
        metadata={
            "description": "Public URI to the tails file",
            "example": (
                "http://192.168.56.133:6543/revocation/registry/"
                f"{ANONCREDS_REV_REG_ID_EXAMPLE}/tails-file"
            ),
        },
    )


class RevRegsCreatedQueryStringSchema(OpenAPISchema):
    """Query string parameters and validators for rev regs created request."""

    cred_def_id = fields.Str(
        required=False,
        validate=ANONCREDS_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )
    state = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(RevRegDefState, m)
                for m in vars(RevRegDefState)
                if m.startswith("STATE_")
            ]
        ),
        metadata={"description": "Revocation registry state"},
    )


class SetRevRegStateQueryStringSchema(OpenAPISchema):
    """Query string parameters and validators for request to set rev reg state."""

    state = fields.Str(
        required=True,
        validate=validate.OneOf(
            [
                getattr(RevRegDefState, m)
                for m in vars(RevRegDefState)
                if m.startswith("STATE_")
            ]
        ),
        metadata={"description": "Revocation registry state to set"},
    )


class RevocationCredDefIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking cred def id."""

    cred_def_id = fields.Str(
        required=True,
        validate=ANONCREDS_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )


class CreateRevRegTxnForEndorserOptionSchema(OpenAPISchema):
    """Class for user to input whether to create a transaction for endorser or not."""

    create_transaction_for_endorser = fields.Boolean(
        required=False,
        metadata={"description": "Create Transaction For Endorser's signature"},
    )


class RevRegConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class PublishRevocationsOptions(OpenAPISchema):
    """Options for publishing revocations to ledger."""

    endorser_connection_id = fields.Str(
        metadata={
            "description": endorser_connection_id_description,
            "required": False,
            "example": UUIDFour.EXAMPLE,
        }
    )

    create_transaction_for_endorser = fields.Bool(
        metadata={
            "description": create_transaction_for_endorser_description,
            "required": False,
            "example": False,
        }
    )


class PublishRevocationsSchemaAnonCreds(OpenAPISchema):
    """Request and result schema for revocation publication API call."""

    rrid2crid = fields.Dict(
        required=False,
        keys=fields.Str(metadata={"example": ANONCREDS_REV_REG_ID_EXAMPLE}),
        values=fields.List(
            fields.Str(
                validate=ANONCREDS_CRED_REV_ID_VALIDATE,
                metadata={
                    "description": "Credential revocation identifier",
                    "example": ANONCREDS_CRED_REV_ID_EXAMPLE,
                },
            )
        ),
        metadata={"description": "Credential revocation ids by revocation registry id"},
    )
    options = fields.Nested(PublishRevocationsOptions())


class PublishRevocationsResultSchemaAnonCreds(OpenAPISchema):
    """Result schema for credential definition send request."""

    rrid2crid = fields.Dict(
        required=False,
        keys=fields.Str(metadata={"example": ANONCREDS_REV_REG_ID_EXAMPLE}),
        values=fields.List(
            fields.Str(
                validate=ANONCREDS_CRED_REV_ID_VALIDATE,
                metadata={
                    "description": "Credential revocation identifier",
                    "example": ANONCREDS_CRED_REV_ID_EXAMPLE,
                },
            )
        ),
        metadata={"description": "Credential revocation ids by revocation registry id"},
    )


class RevokeRequestSchemaAnonCreds(CredRevRecordQueryStringSchema):
    """Parameters and validators for revocation request."""

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate fields - connection_id and thread_id must be present if notify."""
        super().validate_fields(data, **kwargs)

        notify = data.get("notify")
        connection_id = data.get("connection_id")
        notify_version = data.get("notify_version", "v1_0")

        if notify and not connection_id:
            raise ValidationError("Request must specify connection_id if notify is true")
        if notify and not notify_version:
            raise ValidationError("Request must specify notify_version if notify is true")

    publish = fields.Boolean(
        required=False,
        metadata={
            "description": (
                "(True) publish revocation to ledger immediately, or (default, False)"
                " mark it pending"
            )
        },
    )
    notify = fields.Boolean(
        required=False,
        metadata={"description": "Send a notification to the credential recipient"},
    )
    notify_version = fields.String(
        validate=validate.OneOf(["v1_0", "v2_0"]),
        required=False,
        metadata={
            "description": (
                "Specify which version of the revocation notification should be sent"
            )
        },
    )
    connection_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": (
                "Connection ID to which the revocation notification will be sent;"
                " required if notify is true"
            ),
            "example": UUID4_EXAMPLE,
        },
    )
    thread_id = fields.Str(
        required=False,
        metadata={
            "description": (
                "Thread ID of the credential exchange message thread resulting in the"
                " credential now being revoked; required if notify is true"
            )
        },
    )
    comment = fields.Str(
        required=False,
        metadata={
            "description": "Optional comment to include in revocation notification"
        },
    )
    options = PublishRevocationsOptions()


@docs(
    tags=[TAG_TITLE],
    summary="Revoke an issued credential",
)
@request_schema(RevokeRequestSchemaAnonCreds())
@response_schema(AnonCredsRevocationModuleResponseSchema(), description="")
@tenant_authentication
async def revoke(request: web.BaseRequest):
    """Request handler for storing a credential revocation.

    Args:
        request: aiohttp request object

    Returns:
        The credential revocation details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    cred_ex_id = body.get("cred_ex_id")
    body["notify"] = body.get("notify", context.settings.get("revocation.notify"))
    notify = body.get("notify")
    connection_id = body.get("connection_id")
    body["notify_version"] = body.get("notify_version", "v1_0")
    notify_version = body["notify_version"]

    if notify and not connection_id:
        raise web.HTTPBadRequest(reason="connection_id must be set when notify is true")
    if notify and not notify_version:
        raise web.HTTPBadRequest(
            reason="Request must specify notify_version if notify is true"
        )

    rev_manager = RevocationManager(profile)
    try:
        if cred_ex_id:
            # rev_reg_id and cred_rev_id should not be present so we can
            # safely splat the body
            await rev_manager.revoke_credential_by_cred_ex_id(**body)
        else:
            # no cred_ex_id so we can safely splat the body
            await rev_manager.revoke_credential(**body)
        return web.json_response({})
    except (
        RevocationManagerError,
        AnonCredsRevocationError,
        StorageError,
        AnonCredsIssuerError,
        AnonCredsRegistrationError,
    ) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err


@docs(tags=[TAG_TITLE], summary="Publish pending revocations to ledger")
@request_schema(PublishRevocationsSchemaAnonCreds())
@response_schema(PublishRevocationsResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def publish_revocations(request: web.BaseRequest):
    """Request handler for publishing pending revocations to the ledger.

    Args:
        request: aiohttp request object

    Returns:
        Credential revocation ids published as revoked by revocation registry id.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    body = await request.json()
    options = body.get("options", {})
    rrid2crid = body.get("rrid2crid")

    rev_manager = RevocationManager(profile)

    try:
        rev_reg_resp = await rev_manager.publish_pending_revocations(rrid2crid, options)
        return web.json_response({"rrid2crid": rev_reg_resp})
    except (
        RevocationError,
        StorageError,
        AnonCredsIssuerError,
        AnonCredsRevocationError,
    ) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err


@docs(
    tags=[TAG_TITLE],
    summary="Search for matching revocation registries that current agent created",
)
@querystring_schema(RevRegsCreatedQueryStringSchema())
@response_schema(RevRegsCreatedSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def get_rev_regs(request: web.BaseRequest):
    """Request handler to get revocation registries that current agent created.

    Args:
        request: aiohttp request object

    Returns:
        List of identifiers of matching revocation registries.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    search_tags = list(vars(RevRegsCreatedQueryStringSchema)["_declared_fields"])
    tag_filter = {tag: request.query[tag] for tag in search_tags if tag in request.query}
    cred_def_id = tag_filter.get("cred_def_id")
    state = tag_filter.get("state")
    try:
        revocation = AnonCredsRevocation(profile)
        found = await revocation.get_created_revocation_registry_definitions(
            cred_def_id, state
        )
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e
    # TODO remove state == init
    return web.json_response({"rev_reg_ids": found})


@docs(
    tags=[TAG_TITLE],
    summary="Get revocation registry by revocation registry id",
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@response_schema(RevRegResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def get_rev_reg(request: web.BaseRequest):
    """Request handler to get a revocation registry by rev reg id.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    rev_reg = await _get_issuer_rev_reg_record(profile, rev_reg_id)

    return web.json_response({"result": rev_reg.serialize()})


async def _get_issuer_rev_reg_record(
    profile: AskarAnonCredsProfile, rev_reg_id
) -> IssuerRevRegRecord:
    # fetch rev reg def from anoncreds
    try:
        revocation = AnonCredsRevocation(profile)
        rev_reg_def = await revocation.get_created_revocation_registry_definition(
            rev_reg_id
        )
        if rev_reg_def is None:
            raise web.HTTPNotFound(reason="No rev reg def found")
        # looking good, so grab some other data
        state = await revocation.get_created_revocation_registry_definition_state(
            rev_reg_id
        )
        pending_pubs = await revocation.get_pending_revocations(rev_reg_id)
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    # transform
    result = IssuerRevRegRecord(
        record_id=uuid4(),
        state=state,
        cred_def_id=rev_reg_def.cred_def_id,
        error_msg=None,
        issuer_did=rev_reg_def.issuer_id,
        max_cred_num=rev_reg_def.value.max_cred_num,
        revoc_def_type="CL_ACCUM",
        revoc_reg_id=rev_reg_id,
        revoc_reg_def=IndyRevRegDef(
            ver="1.0",
            id_=rev_reg_id,
            revoc_def_type="CL_ACCUM",
            tag=rev_reg_def.tag,
            cred_def_id=rev_reg_def.cred_def_id,
            value=None,
        ),
        revoc_reg_entry=None,
        tag=rev_reg_def.tag,
        tails_hash=rev_reg_def.value.tails_hash,
        tails_local_path=rev_reg_def.value.tails_location,
        tails_public_uri=None,
        pending_pub=pending_pubs,
    )
    return result


@docs(
    tags=[TAG_TITLE],
    summary="Get current active revocation registry by credential definition id",
)
@match_info_schema(RevocationCredDefIdMatchInfoSchema())
@response_schema(RevRegResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def get_active_rev_reg(request: web.BaseRequest):
    """Request handler to get current active revocation registry by cred def id.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry identifier

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    cred_def_id = request.match_info["cred_def_id"]
    try:
        revocation = AnonCredsRevocation(profile)
        active_reg = await revocation.get_or_create_active_registry(cred_def_id)
        rev_reg = await _get_issuer_rev_reg_record(profile, active_reg.rev_reg_def_id)
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    return web.json_response({"result": rev_reg.serialize()})


@docs(tags=[TAG_TITLE], summary="Rotate revocation registry")
@match_info_schema(RevocationCredDefIdMatchInfoSchema())
@response_schema(RevRegsCreatedSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def rotate_rev_reg(request: web.BaseRequest):
    """Request handler to rotate the active revocation registries for cred. def.

    Args:
        request: aiohttp request object

    Returns:
        list or revocation registry ids that were rotated out

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    cred_def_id = request.match_info["cred_def_id"]

    try:
        revocation = AnonCredsRevocation(profile)
        recs = await revocation.decommission_registry(cred_def_id)
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    return web.json_response({"rev_reg_ids": [rec.name for rec in recs if rec.name]})


@docs(
    tags=[TAG_TITLE],
    summary="Get number of credentials issued against revocation registry",
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@response_schema(RevRegIssuedResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def get_rev_reg_issued_count(request: web.BaseRequest):
    """Request handler to get number of credentials issued against revocation registry.

    Args:
        request: aiohttp request object

    Returns:
        Number of credentials issued against revocation registry

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    try:
        revocation = AnonCredsRevocation(profile)
        rev_reg_def = await revocation.get_created_revocation_registry_definition(
            rev_reg_id
        )
        if rev_reg_def is None:
            raise web.HTTPNotFound(reason="No rev reg def found")
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    async with profile.session() as session:
        count = len(
            await IssuerCredRevRecord.query_by_ids(session, rev_reg_id=rev_reg_id)
        )

    return web.json_response({"result": count})


@docs(
    tags=[TAG_TITLE],
    summary="Get details of credentials issued against revocation registry",
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@response_schema(CredRevRecordDetailsResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def get_rev_reg_issued(request: web.BaseRequest):
    """Request handler to get credentials issued against revocation registry.

    Args:
        request: aiohttp request object

    Returns:
        Number of credentials issued against revocation registry

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    try:
        revocation = AnonCredsRevocation(profile)
        rev_reg_def = await revocation.get_created_revocation_registry_definition(
            rev_reg_id
        )
        if rev_reg_def is None:
            raise web.HTTPNotFound(reason="No rev reg def found")
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    async with profile.session() as session:
        recs = await IssuerCredRevRecord.query_by_ids(session, rev_reg_id=rev_reg_id)
    results = []
    for rec in recs:
        results.append(rec.serialize())

    return web.json_response(results)


@docs(
    tags=[TAG_TITLE],
    summary="Get details of revoked credentials from ledger",
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@response_schema(CredRevIndyRecordsResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def get_rev_reg_indy_recs(request: web.BaseRequest):
    """Request handler to get details of revoked credentials from ledger.

    Args:
        request: aiohttp request object

    Returns:
        Details of revoked credentials from ledger

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    indy_registry = LegacyIndyRegistry()

    if await indy_registry.supports(rev_reg_id):
        try:
            rev_reg_delta, _ts = await indy_registry.get_revocation_registry_delta(
                profile, rev_reg_id, None
            )
        except (AnonCredsObjectNotFound, AnonCredsResolutionError) as e:
            raise web.HTTPInternalServerError(reason=str(e)) from e

        return web.json_response(
            {
                "rev_reg_delta": rev_reg_delta,
            }
        )

    raise web.HTTPInternalServerError(
        reason="Indy registry does not support revocation registry "
        f"identified by {rev_reg_id}"
    )


@docs(
    tags=[TAG_TITLE],
    summary="Fix revocation state in wallet and return number of updated entries",
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@querystring_schema(RevRegUpdateRequestMatchInfoSchema())
@response_schema(RevRegWalletUpdatedResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def update_rev_reg_revoked_state(request: web.BaseRequest):
    """Request handler to fix ledger entry of credentials revoked against registry.

    Args:
        request: aiohttp request object

    Returns:
        Number of credentials posted to ledger

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    apply_ledger_update = json.loads(request.query.get("apply_ledger_update", "false"))
    LOGGER.debug(
        "Update revocation state request for rev_reg_id = %s, apply_ledger_update = %s",
        rev_reg_id,
        apply_ledger_update,
    )

    genesis_transactions = None
    recovery_txn = {}
    try:
        revocation = AnonCredsRevocation(profile)
        rev_reg_def = await revocation.get_created_revocation_registry_definition(
            rev_reg_id
        )
        if rev_reg_def is None:
            raise web.HTTPNotFound(reason="No rev reg def found")
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    async with profile.session() as session:
        genesis_transactions = context.settings.get("ledger.genesis_transactions")
        if not genesis_transactions:
            ledger_manager = context.injector.inject(BaseMultipleLedgerManager)
            write_ledger = context.injector.inject(BaseLedger)
            available_write_ledgers = await ledger_manager.get_write_ledgers()
            LOGGER.debug("available write_ledgers = %s", available_write_ledgers)
            LOGGER.debug("write_ledger = %s", write_ledger)
            pool = write_ledger.pool
            LOGGER.debug("write_ledger pool = %s", pool)

            genesis_transactions = pool.genesis_txns

        if not genesis_transactions:
            raise web.HTTPInternalServerError(
                reason="no genesis_transactions for writable ledger"
            )

        if apply_ledger_update:
            ledger = session.inject_or(BaseLedger)
            if not ledger:
                reason = "No ledger available"
                if not session.context.settings.get_value("wallet.type"):
                    reason += ": missing wallet-type?"
                raise web.HTTPInternalServerError(reason=reason)

    rev_manager = RevocationManager(profile)
    try:
        (
            rev_reg_delta,
            recovery_txn,
            applied_txn,
        ) = await rev_manager.update_rev_reg_revoked_state(
            rev_reg_def_id=rev_reg_id,
            apply_ledger_update=apply_ledger_update,
            genesis_transactions=genesis_transactions,
        )
    except (
        RevocationManagerError,
        RevocationError,
        StorageError,
        IndyIssuerError,
        LedgerError,
    ) as err:
        raise web.HTTPBadRequest(reason=err.roll_up)
    except Exception as err:
        raise web.HTTPBadRequest(reason=str(err))

    return web.json_response(
        {
            "rev_reg_delta": rev_reg_delta,
            "recovery_txn": recovery_txn,
            "applied_txn": applied_txn,
        }
    )


@docs(
    tags=[TAG_TITLE],
    summary="Get credential revocation status",
)
@querystring_schema(CredRevRecordQueryStringSchema())
@response_schema(CredRevRecordResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def get_cred_rev_record(request: web.BaseRequest):
    """Request handler to get credential revocation record.

    Args:
        request: aiohttp request object

    Returns:
        The issuer credential revocation record

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.query.get("rev_reg_id")
    cred_rev_id = request.query.get("cred_rev_id")  # numeric string
    cred_ex_id = request.query.get("cred_ex_id")

    try:
        async with profile.session() as session:
            if rev_reg_id and cred_rev_id:
                rec = await IssuerCredRevRecord.retrieve_by_ids(
                    session, rev_reg_id, cred_rev_id
                )
            else:
                rec = await IssuerCredRevRecord.retrieve_by_cred_ex_id(
                    session, cred_ex_id
                )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response({"result": rec.serialize()})


@docs(
    tags=[TAG_TITLE],
    summary="Download tails file",
    produces=["application/octet-stream"],
)
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@response_schema(AnonCredsRevocationModuleResponseSchema, description="tails file")
@tenant_authentication
async def get_tails_file(request: web.BaseRequest) -> web.FileResponse:
    """Request handler to download tails file for revocation registry.

    Args:
        request: aiohttp request object

    Returns:
        The tails file in FileResponse

    """
    #
    # there is no equivalent of this in anoncreds.
    # do we need it there or is this only for transitions.
    #
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    try:
        revocation = AnonCredsRevocation(profile)
        rev_reg_def = await revocation.get_created_revocation_registry_definition(
            rev_reg_id
        )
        if rev_reg_def is None:
            raise web.HTTPNotFound(reason="No rev reg def found")
    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    tails_local_path = rev_reg_def.value.tails_location
    return web.FileResponse(path=tails_local_path, status=200)


@docs(tags=[TAG_TITLE], summary="Set revocation registry state manually")
@match_info_schema(AnonCredsRevRegIdMatchInfoSchema())
@querystring_schema(SetRevRegStateQueryStringSchema())
@response_schema(RevRegResultSchemaAnonCreds(), 200, description="")
@tenant_authentication
async def set_rev_reg_state(request: web.BaseRequest):
    """Request handler to set a revocation registry state manually.

    Args:
        request: aiohttp request object

    Returns:
        The revocation registry record, updated

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_not_anoncreds_profile_raise_web_exception(profile)

    rev_reg_id = request.match_info["rev_reg_id"]
    state = request.query.get("state")

    try:
        revocation = AnonCredsRevocation(profile)
        rev_reg_def = await revocation.set_rev_reg_state(rev_reg_id, state)
        if rev_reg_def is None:
            raise web.HTTPNotFound(reason="No rev reg def found")

    except AnonCredsIssuerError as e:
        raise web.HTTPInternalServerError(reason=str(e)) from e

    rev_reg = await _get_issuer_rev_reg_record(profile, rev_reg_id)
    return web.json_response({"result": rev_reg.serialize()})


class TailsDeleteResponseSchema(OpenAPISchema):
    """Return schema for tails deletion."""

    message = fields.Str()


async def register(app: web.Application):
    """Register routes."""
    app.add_routes(
        [
            web.post("/anoncreds/revocation/revoke", revoke),
            web.post("/anoncreds/revocation/publish-revocations", publish_revocations),
            web.get(
                "/anoncreds/revocation/credential-record",
                get_cred_rev_record,
                allow_head=False,
            ),
            web.get(
                "/anoncreds/revocation/registries",
                get_rev_regs,
                allow_head=False,
            ),
            web.get(
                "/anoncreds/revocation/registry/{rev_reg_id}",
                get_rev_reg,
                allow_head=False,
            ),
            web.get(
                "/anoncreds/revocation/active-registry/{cred_def_id}",
                get_active_rev_reg,
                allow_head=False,
            ),
            web.post(
                "/anoncreds/revocation/active-registry/{cred_def_id}/rotate",
                rotate_rev_reg,
            ),
            web.get(
                "/anoncreds/revocation/registry/{rev_reg_id}/issued",
                get_rev_reg_issued_count,
                allow_head=False,
            ),
            web.get(
                "/anoncreds/revocation/registry/{rev_reg_id}/issued/details",
                get_rev_reg_issued,
                allow_head=False,
            ),
            web.get(
                "/anoncreds/revocation/registry/{rev_reg_id}/issued/indy_recs",
                get_rev_reg_indy_recs,
                allow_head=False,
            ),
            web.get(
                "/anoncreds/revocation/registry/{rev_reg_id}/tails-file",
                get_tails_file,
                allow_head=False,
            ),
            web.patch(
                "/anoncreds/revocation/registry/{rev_reg_id}/set-state",
                set_rev_reg_state,
            ),
            web.put(
                "/anoncreds/revocation/registry/{rev_reg_id}/fix-revocation-entry-state",
                update_rev_reg_revoked_state,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": TAG_TITLE,
            "description": "Revocation registry management",
            "externalDocs": {
                "description": "Overview",
                "url": (
                    "https://github.com/hyperledger/indy-hipe/tree/"
                    "master/text/0011-cred-revocation"
                ),
            },
        }
    )

    # aio_http-apispec polite API only works on schema for JSON objects, not files yet
    methods = app._state["swagger_dict"]["paths"].get(
        "/revocation/registry/{rev_reg_id}/tails-file"
    )
    if methods:
        methods["get"]["responses"]["200"]["schema"] = {
            "type": "string",
            "format": "binary",
        }



"""OOB Problem Report Message Handler."""

from .....messaging.base_handler import BaseHandler
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from ..manager import OutOfBandManager, OutOfBandManagerError
from ..messages.problem_report import OOBProblemReport


class OOBProblemReportMessageHandler(BaseHandler):
    """Handler class for OOB Problem Report Message.

    Updates the ConnRecord Metadata state.
    """

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """OOB Problem Report Message Handler.

        Args:
            context: Request context
            responder: Responder callback
        """
        self._logger.debug(
            "OOBProblemReportMessageHandler called with context %s", context
        )
        assert isinstance(context.message, OOBProblemReport)

        profile = context.profile
        mgr = OutOfBandManager(profile)
        try:
            await mgr.receive_problem_report(
                problem_report=context.message,
                receipt=context.message_receipt,
                conn_record=context.connection_record,
            )
        except OutOfBandManagerError:
            self._logger.exception("Error processing OOB Problem Report message")



"""Web DID Resolver."""

import urllib.parse
from typing import Optional, Pattern, Sequence, Text

import aiohttp
from pydid import DID, DIDDocument

from ...config.injection_context import InjectionContext
from ...core.profile import Profile
from ...messaging.valid import DIDWeb
from ..base import BaseDIDResolver, DIDNotFound, ResolverError, ResolverType


class WebDIDResolver(BaseDIDResolver):
    """Web DID Resolver."""

    def __init__(self):
        """Initialize Web DID Resolver."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for Web DID resolution."""

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported_did_regex of Web DID Resolver."""
        return DIDWeb.PATTERN

    def __transform_to_url(self, did):
        """Transform did to url.

        according to
        https://w3c-ccg.github.io/did-method-web/#read-resolve
        """

        as_did = DID(did)
        method_specific_id = as_did.method_specific_id
        if ":" in method_specific_id:
            # contains path
            url = method_specific_id.replace(":", "/")
        else:
            # bare domain needs /.well-known path
            url = method_specific_id + "/.well-known"

        # Support encoded ports (See: https://github.com/w3c-ccg/did-method-web/issues/7)
        url = urllib.parse.unquote(url)

        return "https://" + url + "/did.json"

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve did:web DIDs."""

        url = self.__transform_to_url(did)
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status == 200:
                    try:
                        # Validate DIDDoc with pyDID
                        did_doc = DIDDocument.from_json(await response.text())
                        return did_doc.serialize()
                    except Exception as err:
                        raise ResolverError("Response was incorrectly formatted") from err
                if response.status == 404:
                    raise DIDNotFound(f"No document found for {did}")
                raise ResolverError(
                    "Could not find doc for {}: {}".format(did, await response.text())
                )



import asyncio
import json

import pytest
from aiohttp.test_utils import AioHTTPTestCase, unused_port

from ....core.profile import Profile
from ....tests import mock
from ....utils.testing import create_test_profile
from ...outbound.message import OutboundMessage
from ...wire_format import JsonWireFormat
from .. import http as test_module
from ..http import HttpTransport
from ..message import InboundMessage
from ..session import InboundSession


class TestHttpTransport(AioHTTPTestCase):
    async def asyncSetUp(self):
        self.message_results = []
        self.port = unused_port()
        self.profile = await create_test_profile()
        self.session = None
        self.transport = HttpTransport(
            "0.0.0.0", self.port, self.create_session, max_message_size=65535
        )
        self.transport.wire_format = JsonWireFormat()
        assert not self.transport.wire_format.get_recipient_keys(None)  # cover method
        self.result_event = None
        self.response_message = None
        await super().asyncSetUp()

    def get_profile(self):
        return self.profile

    def create_session(
        self,
        transport_type,
        *,
        client_info,
        wire_format,
        can_respond: bool = False,
        **kwargs,
    ):
        if not self.session:
            session = InboundSession(
                profile=self.get_profile(),
                can_respond=can_respond,
                inbound_handler=self.receive_message,
                session_id=None,
                wire_format=wire_format,
                client_info=client_info,
                transport_type=transport_type,
            )
            self.session = session
        result = asyncio.Future()
        result.set_result(self.session)
        return result

    def receive_message(
        self,
        profile: Profile,
        message: InboundMessage,
        can_respond: bool = False,
    ):
        message.wait_processing_complete = mock.CoroutineMock()
        self.message_results.append((message.payload, message.receipt, can_respond))
        if self.result_event:
            self.result_event.set()
        if self.response_message and self.session:
            self.session.set_response(self.response_message)

    def get_application(self):
        return self.transport.make_application()

    async def test_start_x(self):
        with mock.patch.object(test_module.web, "TCPSite", mock.MagicMock()) as mock_site:
            mock_site.return_value = mock.MagicMock(
                start=mock.CoroutineMock(side_effect=OSError())
            )
            with pytest.raises(test_module.InboundTransportSetupError):
                await self.transport.start()

    async def test_send_message(self):
        await self.transport.start()

        test_message = {"test": "message"}
        async with self.client.post("/", json=test_message) as resp:
            await resp.text()

        assert self.session is not None
        assert self.session.transport_type == "http"
        assert len(self.message_results) == 1
        assert self.message_results[0][0] == test_message

        await self.transport.stop()

    async def test_send_receive_message(self):
        await self.transport.start()

        test_message = {"~transport": {"return_route": "all"}, "test": "message"}
        test_response = {"response": "ok"}
        self.response_message = OutboundMessage(
            payload=None, enc_payload=json.dumps(test_response)
        )

        async with self.client.post("/", json=test_message) as resp:
            assert await resp.json() == {"response": "ok"}
            # Assert that Server header is cleared
            assert resp.headers.get("Server") is None

        await self.transport.stop()

    async def test_send_message_outliers(self):
        await self.transport.start()

        test_message = {"test": "message"}
        with mock.patch.object(
            test_module.HttpTransport, "create_session", mock.CoroutineMock()
        ) as mock_session:
            mock_session.return_value = mock.MagicMock(
                receive=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        receipt=mock.MagicMock(direct_response_requested=True),
                        wait_processing_complete=mock.CoroutineMock(),
                    )
                ),
                can_respond=True,
                profile=(await create_test_profile()),
                clear_response=mock.MagicMock(),
                wait_response=mock.CoroutineMock(return_value=b"Hello world"),
                response_buffer="something",
            )
            async with self.client.post("/", data=test_message) as resp:
                result = await resp.text()
            assert result == "Hello world"

            mock_session.return_value = mock.MagicMock(
                receive=mock.CoroutineMock(
                    side_effect=test_module.WireFormatParseError()
                ),
                profile=(await create_test_profile()),
            )
            async with self.client.post("/", data=test_message) as resp:
                status = resp.status
                result = await resp.text()
            assert status == 400
            assert str(status) in result

        await self.transport.stop()

    async def test_invite_message_handler(self):
        await self.transport.start()

        request = mock.MagicMock(query={"c_i": "dummy"})
        resp = await self.transport.invite_message_handler(request)
        assert b"You have received a connection invitation" in resp.body
        assert resp.status == 200

        request = mock.MagicMock(query={})
        resp = await self.transport.invite_message_handler(request)
        assert resp.body is None
        assert resp.status == 200

        await self.transport.stop()



from unittest import TestCase, mock

from ......protocols.didcomm_prefix import DIDCommPrefix
from ...message_types import PROTOCOL_PACKAGE, ROTATE
from ...messages.rotate import Rotate

TEST_DID = "test-did"


class TestRotate(TestCase):
    def test_init_type(self):
        """Test initializer."""

        obj = Rotate(to_did=TEST_DID)
        assert obj._type == DIDCommPrefix.qualify_current(ROTATE)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.rotate.RotateSchema.load")
    def test_deserialize(self, mock_rotate_schema_load):
        """Test deserialization."""

        obj = Rotate(to_did=TEST_DID)
        rotate = Rotate.deserialize(obj)
        mock_rotate_schema_load.assert_called_once_with(obj)
        assert rotate is mock_rotate_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.rotate.RotateSchema.dump")
    def test_serialize(self, mock_rotate_schema_dump):
        """Test serialization."""

        obj = Rotate(to_did=TEST_DID)
        rotate_dict = obj.serialize()
        mock_rotate_schema_dump.assert_called_once_with(obj)
        assert rotate_dict is mock_rotate_schema_dump.return_value

    def test_serde(self):
        obj = {
            "to_did": TEST_DID,
        }
        rotate = Rotate.deserialize(obj)
        assert rotate._type == DIDCommPrefix.qualify_current(ROTATE)

        rotate_dict = rotate.serialize()
        assert rotate_dict["to_did"] == obj["to_did"]

    def test_make_model(self):
        """Test making model."""

        obj = Rotate(to_did=TEST_DID)
        data = obj.serialize()
        model_instance = Rotate.deserialize(data)
        assert isinstance(model_instance, Rotate)



from unittest import IsolatedAsyncioTestCase

from ......core.oob_processor import OobMessageProcessor
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.cred_request import V20CredRequest
from ...models.cred_ex_record import V20CredExRecord
from .. import cred_request_handler as test_module

CD_ID = "LjgpST2rjsoxYegQDRm7EL:3:CL:18:tag"


class TestV20CredRequestHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        oob_record = mock.MagicMock()
        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=oob_record
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_cred_mgr.return_value.receive_request.return_value.auto_issue = False
            request_context.message = V20CredRequest()
            request_context.connection_ready = True
            handler = test_module.V20CredRequestHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_request.assert_called_once_with(
            request_context.message, request_context.connection_record, oob_record
        )
        assert not responder.messages

    async def test_called_auto_issue(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        oob_record = mock.MagicMock()
        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=oob_record
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        cred_ex_rec = V20CredExRecord()

        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=cred_ex_rec
            )
            mock_cred_mgr.return_value.receive_request.return_value.auto_issue = True
            mock_cred_mgr.return_value.issue_credential = mock.CoroutineMock(
                return_value=(None, "cred_issue_message")
            )
            request_context.message = V20CredRequest()
            request_context.connection_ready = True
            handler = test_module.V20CredRequestHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)
            mock_cred_mgr.return_value.issue_credential.assert_called_once_with(
                cred_ex_record=cred_ex_rec, comment=None
            )

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_request.assert_called_once_with(
            request_context.message, request_context.connection_record, oob_record
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "cred_issue_message"
        assert target == {}

    async def test_called_auto_issue_x_indy(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        cred_ex_rec = V20CredExRecord()

        oob_record = mock.MagicMock()
        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=oob_record
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with (
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(cred_ex_rec, "save_error_state", mock.CoroutineMock()),
        ):
            mock_cred_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=cred_ex_rec
            )
            mock_cred_mgr.return_value.receive_request.return_value.auto_issue = True
            mock_cred_mgr.return_value.issue_credential = mock.CoroutineMock(
                side_effect=test_module.IndyIssuerError()
            )

            request_context.message = V20CredRequest()
            request_context.connection_ready = True
            handler = test_module.V20CredRequestHandler()
            responder = MockResponder()

            with (
                mock.patch.object(
                    responder, "send_reply", mock.CoroutineMock()
                ) as mock_send_reply,
                mock.patch.object(
                    handler._logger, "exception", mock.MagicMock()
                ) as mock_log_exc,
            ):
                await handler.handle(request_context, responder)
                mock_log_exc.assert_called_once()

    async def test_called_auto_issue_x_anoncreds(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        cred_ex_rec = V20CredExRecord()

        oob_record = mock.MagicMock()
        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=oob_record
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with (
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(cred_ex_rec, "save_error_state", mock.CoroutineMock()),
        ):
            mock_cred_mgr.return_value.receive_request = mock.CoroutineMock(
                return_value=cred_ex_rec
            )
            mock_cred_mgr.return_value.receive_request.return_value.auto_issue = True
            mock_cred_mgr.return_value.issue_credential = mock.AsyncMock(
                side_effect=test_module.AnonCredsIssuerError()
            )

            request_context.message = V20CredRequest()
            request_context.connection_ready = True
            handler = test_module.V20CredRequestHandler()
            responder = MockResponder()

            with (
                mock.patch.object(
                    responder, "send_reply", mock.CoroutineMock()
                ) as mock_send_reply,
                mock.patch.object(
                    handler._logger, "exception", mock.MagicMock()
                ) as mock_log_exc,
            ):
                await handler.handle(request_context, responder)
                mock_log_exc.assert_called_once()

    async def test_called_not_ready(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_request = mock.CoroutineMock()
            request_context.message = V20CredRequest()
            request_context.connection_ready = False
            handler = test_module.V20CredRequestHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler.handle(request_context, responder)
            assert (
                err.exception.message
                == "Connection used for credential request not ready"
            )

        assert not responder.messages

    async def test_called_no_connection_no_oob(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()

        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=None
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        request_context.message = V20CredRequest()
        handler = test_module.V20CredRequestHandler()
        responder = MockResponder()
        with self.assertRaises(test_module.HandlerException) as err:
            await handler.handle(request_context, responder)

        assert not responder.messages



"""the did resolver.

responsible for keeping track of all resolvers. more importantly
retrieving did's from different sources provided by the method type.
"""

import asyncio
import logging
from datetime import datetime, timezone
from itertools import chain
from typing import List, Optional, Sequence, Text, Tuple, Union

import pydid
from pydid import DID, DIDError, DIDUrl, Resource, VerificationMethod
from pydid.doc.doc import BaseDIDDocument, IDNotFoundError

from ..core.profile import Profile
from .base import (
    BaseDIDResolver,
    DIDMethodNotSupported,
    DIDNotFound,
    ResolutionMetadata,
    ResolutionResult,
    ResolverError,
)

LOGGER = logging.getLogger(__name__)


class DIDResolver:
    """did resolver singleton."""

    DEFAULT_TIMEOUT = 30

    def __init__(self, resolvers: Optional[List[BaseDIDResolver]] = None):
        """Create DID Resolver."""
        self.resolvers = resolvers or []

    def register_resolver(self, resolver: BaseDIDResolver):
        """Register a new resolver."""
        self.resolvers.append(resolver)

    async def _resolve(
        self,
        profile: Profile,
        did: Union[str, DID],
        service_accept: Optional[Sequence[Text]] = None,
        *,
        timeout: Optional[int] = None,
    ) -> Tuple[BaseDIDResolver, dict]:
        """Retrieve doc and return with resolver.

        This private method enables the public resolve and resolve_with_metadata
        methods to share the same logic.
        """
        if isinstance(did, DID):
            did = str(did)
        else:
            DID.validate(did)
        for resolver in await self._match_did_to_resolver(profile, did):
            try:
                LOGGER.debug("Resolving DID %s with %s", did, resolver)
                document = await asyncio.wait_for(
                    resolver.resolve(
                        profile,
                        did,
                        service_accept,
                    ),
                    timeout if timeout is not None else self.DEFAULT_TIMEOUT,
                )
                LOGGER.debug("Resolved DID %s with %s: %s", did, resolver, document)
                return resolver, document
            except DIDNotFound:
                LOGGER.debug("DID %s not found by resolver %s", did, resolver)

        raise DIDNotFound(f"DID {did} could not be resolved")

    async def resolve(
        self,
        profile: Profile,
        did: Union[str, DID],
        service_accept: Optional[Sequence[Text]] = None,
        *,
        timeout: Optional[int] = None,
    ) -> dict:
        """Resolve a DID."""
        _, doc = await self._resolve(profile, did, service_accept, timeout=timeout)
        return doc

    async def resolve_with_metadata(
        self, profile: Profile, did: Union[str, DID], *, timeout: Optional[int] = None
    ) -> ResolutionResult:
        """Resolve a DID and return the ResolutionResult."""
        resolution_start_time = datetime.now(tz=timezone.utc)

        resolver, doc = await self._resolve(profile, did, timeout=timeout)

        time_now = datetime.now(tz=timezone.utc)
        duration = int((time_now - resolution_start_time).total_seconds() * 1000)
        retrieved_time = time_now.strftime("%Y-%m-%dT%H:%M:%SZ")
        resolver_metadata = ResolutionMetadata(
            resolver.type, type(resolver).__qualname__, retrieved_time, duration
        )
        return ResolutionResult(doc, resolver_metadata)

    async def _match_did_to_resolver(
        self, profile: Profile, did: str
    ) -> Sequence[BaseDIDResolver]:
        """Generate supported DID Resolvers.

        Native resolvers are yielded first, in registered order followed by
        non-native resolvers in registered order.
        """
        valid_resolvers = [
            resolver
            for resolver in self.resolvers
            if await resolver.supports(profile, did)
        ]
        LOGGER.debug("Valid resolvers for DID %s: %s", did, valid_resolvers)
        native_resolvers = filter(lambda resolver: resolver.native, valid_resolvers)
        non_native_resolvers = filter(
            lambda resolver: not resolver.native, valid_resolvers
        )
        resolvers = list(chain(native_resolvers, non_native_resolvers))
        if not resolvers:
            raise DIDMethodNotSupported(f'No resolver supporting DID "{did}" loaded')
        return resolvers

    async def dereference(
        self,
        profile: Profile,
        did_url: str,
        *,
        document: Optional[BaseDIDDocument] = None,
    ) -> Resource:
        """Dereference a DID URL to its corresponding DID Doc object."""
        try:
            parsed = DIDUrl.parse(did_url)
            if not parsed.did and not document:
                raise ValueError("Invalid DID URL")
        except DIDError as err:
            raise ResolverError(
                "Failed to parse DID URL from {}".format(did_url)
            ) from err

        if document and parsed.did and parsed.did != document.id:
            document = None

        if not document:
            doc_dict = await self.resolve(profile, parsed.did)
            document = pydid.deserialize_document(doc_dict)

        try:
            return document.dereference(parsed)
        except IDNotFoundError as error:
            raise ResolverError(
                "Failed to dereference DID URL: {}".format(error)
            ) from error

    async def dereference_verification_method(
        self,
        profile: Profile,
        did_url: str,
        *,
        document: Optional[BaseDIDDocument] = None,
    ) -> VerificationMethod:
        """Dereference a DID URL to a verification method."""
        dereferenced = await self.dereference(profile, did_url, document=document)
        if isinstance(dereferenced, VerificationMethod):
            return dereferenced
        raise ValueError("DID URL does not dereference to a verification method")



from unittest import IsolatedAsyncioTestCase

from ...protocols.issue_credential.v1_0.message_types import CONTROLLERS
from ..goal_code_registry import GoalCodeRegistry


class TestGoalCodeRegistry(IsolatedAsyncioTestCase):
    test_goal_code_queries = "*"
    test_goal_code_queries_fail = "aries.fake.*"

    def setUp(self):
        self.registry = GoalCodeRegistry()

    def test_goal_codes(self):
        self.registry.register_controllers(CONTROLLERS)
        matches = self.registry.goal_codes_matching_query(self.test_goal_code_queries)
        assert tuple(matches) == ("aries.vc", "aries.vc.issue")
        matches = self.registry.goal_codes_matching_query(
            self.test_goal_code_queries_fail
        )
        assert tuple(matches) == ()



import pytest
import pytest_asyncio

from ......core.protocol_registry import ProtocolRegistry
from ......messaging.base_handler import HandlerException
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......utils.testing import create_test_profile
from .....didcomm_prefix import DIDCommPrefix
from ...handlers.disclose_handler import DiscloseHandler
from ...messages.disclose import Disclose
from ...messages.query import Query
from ...models.discovery_record import V10DiscoveryExchangeRecord

TEST_MESSAGE_FAMILY = "doc/proto/1.0"
TEST_MESSAGE_TYPE = TEST_MESSAGE_FAMILY + "/message"


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    ctx.connection_ready = True
    ctx.connection_record = mock.MagicMock(connection_id="test123")
    yield ctx


class TestDiscloseHandler:
    @pytest.mark.asyncio
    async def test_disclose(self, request_context: RequestContext):
        registry = ProtocolRegistry()
        registry.register_message_types({TEST_MESSAGE_TYPE: object()})
        request_context.injector.bind_instance(ProtocolRegistry, registry)
        disclose_msg = Disclose(
            protocols=[
                {
                    "pid": DIDCommPrefix.qualify_current("test_proto/v1.0/test_message"),
                    "roles": [],
                }
            ]
        )
        query_msg = Query(query="*")
        discovery_record = V10DiscoveryExchangeRecord(
            connection_id="test123",
            thread_id="test123",
            query_msg=query_msg,
        )
        disclose_msg.assign_thread_id("test123")
        request_context.message = disclose_msg

        handler = DiscloseHandler()
        mock_responder = MockResponder()
        with mock.patch.object(
            V10DiscoveryExchangeRecord,
            "retrieve_by_id",
            mock.CoroutineMock(return_value=discovery_record),
        ):
            await handler.handle(request_context, mock_responder)
            assert not mock_responder.messages

    @pytest.mark.asyncio
    async def test_disclose_connection_not_ready(self, request_context):
        request_context.connection_ready = False
        disclose_msg = Disclose(
            protocols=[
                {
                    "pid": DIDCommPrefix.qualify_current("test_proto/v1.0/test_message"),
                    "roles": [],
                }
            ]
        )
        disclose_msg.assign_thread_id("test123")
        request_context.message = disclose_msg

        handler = DiscloseHandler()
        mock_responder = MockResponder()
        with pytest.raises(HandlerException):
            await handler.handle(request_context, mock_responder)



from .....messaging.valid import UUID4_EXAMPLE

test_conn_id = UUID4_EXAMPLE


class MockConnRecord:
    def __init__(self, connection_id, is_ready) -> None:
        self.connection_id = connection_id
        self.is_ready = is_ready



from copy import deepcopy
from unittest import IsolatedAsyncioTestCase

import pytest

from ...anoncreds.models.credential_definition import (
    CredDef,
    CredDefValue,
    CredDefValuePrimary,
    CredDefValueRevocation,
    GetCredDefResult,
)
from ...anoncreds.models.revocation import (
    GetRevListResult,
    GetRevRegDefResult,
    RevList,
    RevRegDef,
    RevRegDefValue,
)
from ...anoncreds.models.schema import AnonCredsSchema, GetSchemaResult
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import verifier as test_module
from .mock_objects import (
    MOCK_CRED_DEFS,
    MOCK_PRES,
    MOCK_PRES_REQ,
    MOCK_REV_REG_DEFS,
    MOCK_SCHEMAS,
)


@pytest.mark.anoncreds
class TestAnonCredsVerifier(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(
            settings={"wallet.type": "askar-anoncreds"},
        )
        self.verifier = test_module.AnonCredsVerifier(self.profile)

    async def test_init(self):
        assert self.verifier.profile == self.profile

    async def test_non_revoc_intervals(self):
        result = self.verifier.non_revoc_intervals(
            MOCK_PRES_REQ, MOCK_PRES, MOCK_CRED_DEFS
        )
        assert isinstance(result, list)

        non_revoked_req = deepcopy(MOCK_PRES_REQ)
        non_revoked_req["requested_attributes"]["biomarker_attrs_0"]["non_revoked"] = (
            {
                "from": 1593922800,
                "to": 1593922800,
            },
        )

        self.verifier.non_revoc_intervals(non_revoked_req, MOCK_PRES, MOCK_CRED_DEFS)

    async def test_check_timestamps_with_names(self):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="TUku9MDGa7QALbAJX4oAww:3:CL:531757:MYCO_Consent_Enablement",
                        credential_definition=CredDef(
                            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )
        await self.verifier.check_timestamps(
            self.profile, MOCK_PRES_REQ, MOCK_PRES, MOCK_REV_REG_DEFS
        )

        # irrevocable cred-def with timestamp
        mock_pres = deepcopy(MOCK_PRES)
        mock_pres["identifiers"][0]["timestamp"] = 9999999999
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, MOCK_PRES_REQ, mock_pres, MOCK_REV_REG_DEFS
            )

        # revocable cred-def with timestamp
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="TUku9MDGa7QALbAJX4oAww:3:CL:531757:MYCO_Consent_Enablement",
                        credential_definition=CredDef(
                            issuer_id="issuer-id",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z"),
                                revocation=CredDefValueRevocation(
                                    g="g",
                                    g_dash="g_dash",
                                    h="h",
                                    h0="h0",
                                    h1="h1",
                                    h2="h2",
                                    h_cap="h_cap",
                                    htilde="htilde",
                                    pk="pk",
                                    u="u",
                                    y="y",
                                ),
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )

        # too far in future
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, MOCK_PRES_REQ, mock_pres, MOCK_REV_REG_DEFS
            )

        # no rev_reg_id
        mock_pres["identifiers"][0]["timestamp"] = 1234567890
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, MOCK_PRES_REQ, mock_pres, MOCK_REV_REG_DEFS
            )

        # with rev_reg_id
        mock_pres["identifiers"][0]["rev_reg_id"] = (
            "TUku9MDGa7QALbAJX4oAww:3:TUku9MDGa7QALbAJX4oAww:3:CL:18:tag:CL_ACCUM:0"
        )

        # Superfluous timestamp
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, MOCK_PRES_REQ, mock_pres, MOCK_REV_REG_DEFS
            )

        # Valid
        mock_pres["identifiers"][0]["timestamp"] = None
        await self.verifier.check_timestamps(
            self.profile, MOCK_PRES_REQ, mock_pres, MOCK_REV_REG_DEFS
        )

        #  outside of non-revocation interval
        mock_pres_req = deepcopy(MOCK_PRES_REQ)
        mock_pres_req["requested_attributes"]["biomarker_attrs_0"]["non_revoked"] = {
            "from": 9000000000,
            "to": 1000000000,
        }
        mock_pres["identifiers"][0]["timestamp"] = 123456789
        await self.verifier.check_timestamps(
            self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
        )

        # no revealed attr groups
        mock_pres["requested_proof"]["revealed_attr_groups"] = {}
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
            )

    async def test_check_timestamps_with_name(self):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="TUku9MDGa7QALbAJX4oAww:3:CL:531757:MYCO_Consent_Enablement",
                        credential_definition=CredDef(
                            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )
        mock_pres_req = deepcopy(MOCK_PRES_REQ)
        mock_pres = deepcopy(MOCK_PRES)
        del mock_pres_req["requested_attributes"]["biomarker_attrs_0"]["names"]
        mock_pres_req["requested_attributes"]["biomarker_attrs_0"]["name"] = "name"

        # invalid group
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
            )

        # valid
        del mock_pres["requested_proof"]["revealed_attrs"]["consent_attrs"]
        del mock_pres["requested_proof"]["revealed_attr_groups"]["biomarker_attrs_0"]
        del mock_pres_req["requested_attributes"]["consent_attrs"]
        mock_pres["requested_proof"]["revealed_attrs"]["biomarker_attrs_0"] = {
            "sub_proof_index": 0,
            "raw": "Iron",
            "encoded": "85547618788485118809771015708850341281587970912661276233439574555663751388073",
        }

        await self.verifier.check_timestamps(
            self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
        )

        # revocable cred-def with timestamp
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="TUku9MDGa7QALbAJX4oAww:3:CL:531757:MYCO_Consent_Enablement",
                        credential_definition=CredDef(
                            issuer_id="issuer-id",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z"),
                                revocation=CredDefValueRevocation(
                                    g="g",
                                    g_dash="g_dash",
                                    h="h",
                                    h0="h0",
                                    h1="h1",
                                    h2="h2",
                                    h_cap="h_cap",
                                    htilde="htilde",
                                    pk="pk",
                                    u="u",
                                    y="y",
                                ),
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )

        # no rev_reg_id
        mock_pres["identifiers"][0]["timestamp"] = 1234567890
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
            )

        # with rev_reg_id
        mock_pres["identifiers"][0]["rev_reg_id"] = (
            "TUku9MDGa7QALbAJX4oAww:3:TUku9MDGa7QALbAJX4oAww:3:CL:18:tag:CL_ACCUM:0"
        )

        # Superfluous timestamp
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
            )

        #  outside of non-revocation interval
        mock_pres_req["requested_attributes"]["biomarker_attrs_0"]["non_revoked"] = {
            "from": 9000000000,
            "to": 1000000000,
        }
        mock_pres["identifiers"][0]["timestamp"] = 123456789
        await self.verifier.check_timestamps(
            self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
        )

        # unrevealed attr
        mock_pres["requested_proof"]["revealed_attrs"] = {}
        mock_pres["requested_proof"]["unrevealed_attrs"]["biomarker_attrs_0"] = {
            "sub_proof_index": 0,
        }
        await self.verifier.check_timestamps(
            self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
        )

    async def test_check_timestamps_predicates(self):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="TUku9MDGa7QALbAJX4oAww:3:CL:531757:MYCO_Consent_Enablement",
                        credential_definition=CredDef(
                            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )

        mock_pres_req = deepcopy(MOCK_PRES_REQ)
        mock_pres_req["requested_attributes"] = {}
        mock_pres_req["requested_predicates"]["0_concentration_GE_uuid"] = {
            "name": "concentration",
            "p_type": "<=",
            "p_value": 9,
            "restrictions": [{"schema_name": "MYCO Biomarker"}],
        }

        # predicate not in proof
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, mock_pres_req, MOCK_PRES, MOCK_REV_REG_DEFS
            )

        mock_pres = deepcopy(MOCK_PRES)
        mock_pres["requested_proof"]["predicates"] = {
            "0_concentration_GE_uuid": {"sub_proof_index": 0}
        }

        #  valid - no revocation
        await self.verifier.check_timestamps(
            self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
        )

        # revocation

        # revocable cred-def with timestamp
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="TUku9MDGa7QALbAJX4oAww:3:CL:531757:MYCO_Consent_Enablement",
                        credential_definition=CredDef(
                            issuer_id="issuer-id",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z"),
                                revocation=CredDefValueRevocation(
                                    g="g",
                                    g_dash="g_dash",
                                    h="h",
                                    h0="h0",
                                    h1="h1",
                                    h2="h2",
                                    h_cap="h_cap",
                                    htilde="htilde",
                                    pk="pk",
                                    u="u",
                                    y="y",
                                ),
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )

        # no rev_reg_id
        mock_pres["identifiers"][0]["timestamp"] = 1234567890
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
            )

        # with rev_reg_id
        mock_pres["identifiers"][0]["rev_reg_id"] = (
            "TUku9MDGa7QALbAJX4oAww:3:TUku9MDGa7QALbAJX4oAww:3:CL:18:tag:CL_ACCUM:0"
        )

        # Superfluous timestamp
        with self.assertRaises(ValueError):
            await self.verifier.check_timestamps(
                self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
            )

        #  outside of non-revocation interval
        mock_pres_req["requested_predicates"]["0_concentration_GE_uuid"][
            "non_revoked"
        ] = {
            "from": 9000000000,
            "to": 1000000000,
        }
        mock_pres["identifiers"][0]["timestamp"] = 123456789
        await self.verifier.check_timestamps(
            self.profile, mock_pres_req, mock_pres, MOCK_REV_REG_DEFS
        )

    async def test_pre_verify_incomplete_objects(self):
        mock_pres_req = deepcopy(MOCK_PRES_REQ)
        mock_pres = deepcopy(MOCK_PRES)

        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, {})

        with self.assertRaises(ValueError):
            await self.verifier.pre_verify({}, mock_pres)

        del mock_pres_req["requested_predicates"]
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)
        mock_pres_req["requested_predicates"] = {}

        del mock_pres_req["requested_attributes"]
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)
        mock_pres_req["requested_attributes"] = {}

        del mock_pres["requested_proof"]
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)
        mock_pres["requested_proof"] = {}

        del mock_pres["proof"]
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)

    async def test_pre_verify(self):
        mock_pres_req = deepcopy(MOCK_PRES_REQ)
        mock_pres = deepcopy(MOCK_PRES)

        # valid - with names and name
        result = await self.verifier.pre_verify(mock_pres_req, mock_pres)
        assert isinstance(result, list)

        # name in unrevealed attrs
        mock_pres["requested_proof"]["revealed_attrs"] = {}
        mock_pres["requested_proof"]["unrevealed_attrs"] = {
            "consent_attrs": {"sub_proof_index": 1}
        }
        await self.verifier.pre_verify(mock_pres_req, mock_pres)

        # self-attested attrs with restrictions
        mock_pres["requested_proof"]["unrevealed_attrs"] = {}
        mock_pres["requested_proof"]["self_attested_attrs"] = {
            "consent_attrs": "I agree to share my data with the verifier"
        }
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)

        # valid - self-attested
        mock_pres_req["requested_attributes"]["consent_attrs"]["restrictions"] = []
        await self.verifier.pre_verify(mock_pres_req, mock_pres)

        #  no name or names
        del mock_pres_req["requested_attributes"]["consent_attrs"]["name"]
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)
        mock_pres_req["requested_attributes"]["consent_attrs"]["name"] = (
            "jti_unique_identifier"
        )
        # attr not in proof
        mock_pres["requested_proof"]["self_attested_attrs"] = {}
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)

        # predicates

        # predicate not in proof
        mock_pres_req["requested_attributes"] = {}
        mock_pres_req["requested_predicates"]["0_concentration_GE_uuid"] = {
            "name": "concentration",
            "p_type": "<=",
            "p_value": 9,
            "restrictions": [{"schema_name": "MYCO Biomarker"}],
        }

        mock_pres["requested_proof"]["predicates"] = {
            "0_concentration_GE_uuid": {"sub_proof_index": 0}
        }
        with self.assertRaises(ValueError):
            await self.verifier.pre_verify(mock_pres_req, mock_pres)

        mock_pres["proof"]["proofs"][0]["primary_proof"]["ge_proofs"] = [
            {
                "predicate": {
                    "attr_name": "concentration",
                    "p_type": "<=",
                    "value": 9,
                },
            }
        ]
        await self.verifier.pre_verify(mock_pres_req, mock_pres)

    async def test_process_pres_identifiers(self):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(
                    return_value=GetSchemaResult(
                        schema_id="schema-id",
                        schema=AnonCredsSchema(
                            issuer_id="issuer-id",
                            name="schema-name",
                            version="1.0",
                            attr_names=["attr1", "attr2"],
                        ),
                        schema_metadata={},
                        resolution_metadata={},
                    )
                ),
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="cred-def-id",
                        credential_definition=CredDef(
                            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                ),
                get_revocation_registry_definition=mock.CoroutineMock(
                    return_value=GetRevRegDefResult(
                        revocation_registry_id="rev-reg-id",
                        revocation_registry=RevRegDef(
                            issuer_id="issuer-id",
                            cred_def_id="cred-def-id",
                            type="CL_ACCUM",
                            tag="tag",
                            value=RevRegDefValue(
                                public_keys={},
                                max_cred_num=1000,
                                tails_hash="tails-hash",
                                tails_location="tails-location",
                            ),
                        ),
                        resolution_metadata={},
                        revocation_registry_metadata={},
                    )
                ),
                get_revocation_list=mock.CoroutineMock(
                    return_value=GetRevListResult(
                        revocation_list=RevList(
                            issuer_id="issuer-id",
                            rev_reg_def_id="rev-reg-def-id",
                            revocation_list=[],
                            current_accumulator="current-accumulator",
                        ),
                        resolution_metadata={},
                        revocation_registry_metadata={},
                    )
                ),
            )
        )
        result = await self.verifier.process_pres_identifiers(
            [
                {
                    "schema_id": "schema-id",
                    "cred_def_id": "cred-def-id",
                    "rev_reg_id": "rev-reg-id",
                    "timestamp": 1234567890,
                }
            ]
        )

        assert isinstance(result, tuple)
        assert len(result) == 4

    async def test_verify_presentation(self):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    return_value=GetCredDefResult(
                        credential_definition_id="cred-def-id",
                        credential_definition=CredDef(
                            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
                            schema_id="schema-id",
                            tag="tag",
                            type="CL",
                            value=CredDefValue(
                                primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                            ),
                        ),
                        credential_definition_metadata={},
                        resolution_metadata={},
                    )
                )
            )
        )
        result = await self.verifier.verify_presentation(
            pres_req=MOCK_PRES_REQ,
            pres=MOCK_PRES,
            schemas=MOCK_SCHEMAS,
            credential_definitions=MOCK_CRED_DEFS,
            rev_reg_defs=[],
            rev_lists={},
        )

        assert isinstance(result, tuple)
        assert len(result) == 2
        assert result[0] is False

    @mock.patch.object(
        test_module.AnonCredsVerifier, "pre_verify", side_effect=ValueError()
    )
    async def test_verify_presentation_value_error_caught(self, mock_verify):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    side_effect=ValueError("Bad credential definition")
                )
            )
        )
        (result, msgs) = await self.verifier.verify_presentation(
            pres_req=MOCK_PRES_REQ,
            pres=MOCK_PRES,
            schemas=MOCK_SCHEMAS,
            credential_definitions=MOCK_CRED_DEFS,
            rev_reg_defs=[],
            rev_lists={},
        )
        assert result is False
        assert isinstance(msgs, list)



from unittest import IsolatedAsyncioTestCase

from ...core.error import StartupError
from ...core.profile import ProfileManager, ProfileSession
from ...storage.base import BaseStorage
from ...storage.record import StorageRecord
from ...tests import mock
from ...utils.testing import create_test_profile
from ...version import RECORD_TYPE_ACAPY_VERSION, __version__
from ...wallet.base import BaseWallet
from .. import wallet as test_module
from ..injection_context import InjectionContext
from ..injector import Injector

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"


class MockManager(ProfileManager):
    def __init__(self, profile):
        self.profile = profile
        self.calls = []

    async def provision(self, context, config):
        self.calls.append(("provision", context, config))
        return self.profile

    async def open(self, context, config):
        self.calls.append(("open", context, config))
        return self.profile


class TestWalletConfig(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.injector = Injector(enforce_typing=False)
        self.session = mock.MagicMock(ProfileSession)()
        self.session.commit = mock.CoroutineMock()
        self.profile = mock.MagicMock(
            backend="indy",
            created=True,
            name="Test Wallet",
            transaction=mock.CoroutineMock(return_value=self.session),
        )

        def _inject(cls):
            return self.injector.inject(cls)

        self.session.inject = _inject
        self.context = InjectionContext()
        self.context.injector.bind_instance(ProfileManager, MockManager(self.profile))

    async def test_wallet_config_existing_replace(self):
        self.context.update_settings(
            {
                "wallet.seed": "00000000000000000000000000000000",
                "wallet.replace_public_did": True,
                "debug.enabled": True,
            }
        )
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
            set_public_did=mock.CoroutineMock(),
            create_local_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        with (
            mock.patch.object(
                test_module, "seed_to_did", mock.MagicMock()
            ) as mock_seed_to_did,
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            mock_seed_to_did.return_value = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"

            await test_module.wallet_config(self.context, provision=True)

    async def test_wallet_config_existing_open(self):
        self.profile = mock.MagicMock(
            backend="indy",
            created=False,
            name="Test Wallet",
            transaction=mock.CoroutineMock(return_value=self.session),
        )
        self.context.injector.clear_binding(ProfileManager)
        self.context.injector.bind_instance(ProfileManager, MockManager(self.profile))

        self.context.update_settings(
            {
                "wallet.seed": "00000000000000000000000000000000",
                "wallet.replace_public_did": True,
                "debug.enabled": True,
            }
        )
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
            set_public_did=mock.CoroutineMock(),
            create_local_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        self.context.injector.bind_instance(ProfileManager, MockManager(self.profile))

        with (
            mock.patch.object(
                test_module, "seed_to_did", mock.MagicMock()
            ) as mock_seed_to_did,
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            mock_seed_to_did.return_value = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"

            await test_module.wallet_config(self.context, provision=True)

    async def test_wallet_config_auto_provision(self):
        self.context.update_settings(
            {
                "wallet.seed": "00000000000000000000000000000000",
                "wallet.replace_public_did": True,
                "debug.enabled": True,
                "auto_provision": False,
            }
        )
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
            set_public_did=mock.CoroutineMock(),
            create_local_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        with (
            mock.patch.object(MockManager, "open", mock.CoroutineMock()) as mock_mgr_open,
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            mock_mgr_open.side_effect = test_module.ProfileNotFoundError()

            with self.assertRaises(StartupError):
                await test_module.wallet_config(self.context, provision=False)

            self.context.update_settings({"auto_provision": True})
            profile, did_info = await test_module.wallet_config(
                self.context, provision=False
            )

            self.assertEqual(profile, self.profile)
            self.assertIsNotNone(did_info)
            self.assertEqual(did_info.did, TEST_DID)
            self.assertEqual(did_info.verkey, TEST_VERKEY)

    async def test_wallet_config_non_indy_x(self):
        self.context.update_settings(
            {
                "wallet.seed": "00000000000000000000000000000000",
            }
        )
        self.profile.backend = "not-indy"
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)
        with mock.patch.object(
            test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
        ):
            with self.assertRaises(test_module.ConfigError):
                await test_module.wallet_config(self.context, provision=True)

    async def test_wallet_config_bad_seed_x(self):
        self.context.update_settings(
            {
                "wallet.seed": "00000000000000000000000000000000",
            }
        )
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        with (
            mock.patch.object(
                test_module,
                "seed_to_did",
                mock.MagicMock(return_value="XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"),
            ),
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            with self.assertRaises(test_module.ConfigError):
                await test_module.wallet_config(self.context, provision=True)

    async def test_wallet_config_seed_local(self):
        self.context.update_settings(
            {
                "wallet.seed": "00000000000000000000000000000000",
                "wallet.local_did": True,
            }
        )
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(return_value=None),
            set_public_did=mock.CoroutineMock(),
            create_local_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        with (
            mock.patch.object(
                test_module, "seed_to_did", mock.MagicMock()
            ) as mock_seed_to_did,
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            mock_seed_to_did.return_value = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"

            profile, did_info = await test_module.wallet_config(
                self.context, provision=True
            )

            self.assertEqual(profile, self.profile)
            self.assertIsNotNone(did_info)
            self.assertEqual(did_info.did, TEST_DID)
            self.assertEqual(did_info.verkey, TEST_VERKEY)

    async def test_wallet_config_seed_public(self):
        self.context.update_settings(
            {
                "wallet.seed": "00000000000000000000000000000000",
            }
        )
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(return_value=None),
            set_public_did=mock.CoroutineMock(),
            create_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        with (
            mock.patch.object(
                test_module,
                "seed_to_did",
                mock.MagicMock(return_value="XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"),
            ),
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            profile, did_info = await test_module.wallet_config(
                self.context, provision=True
            )

            self.assertEqual(profile, self.profile)
            self.assertIsNotNone(did_info)
            self.assertEqual(did_info.did, TEST_DID)
            self.assertEqual(did_info.verkey, TEST_VERKEY)

    async def test_wallet_config_seed_no_public_did(self):
        self.context.update_settings({"wallet.seed": "original_seed"})
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(return_value=None),
            set_public_did=mock.CoroutineMock(),
            create_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        with (
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            profile, did_info = await test_module.wallet_config(
                self.context, provision=True
            )

            self.assertEqual(profile, self.profile)
            self.assertIsNotNone(did_info)
            self.assertEqual(did_info.did, TEST_DID)
            self.assertEqual(did_info.verkey, TEST_VERKEY)

    async def test_wallet_config_for_key_derivation_method(self):
        self.context.update_settings(
            {
                "wallet.key_derivation_method": "derivation_method",
            }
        )
        mock_wallet = mock.MagicMock(
            get_public_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
            set_public_did=mock.CoroutineMock(),
            create_local_did=mock.CoroutineMock(
                return_value=mock.MagicMock(did=TEST_DID, verkey=TEST_VERKEY)
            ),
        )
        self.injector.bind_instance(BaseWallet, mock_wallet)

        with (
            mock.patch.object(
                MockManager, "provision", mock.CoroutineMock()
            ) as mock_mgr_provision,
            mock.patch.object(
                test_module, "add_or_update_version_to_storage", mock.CoroutineMock()
            ),
        ):
            mock_mgr_provision.return_value = self.profile

            await test_module.wallet_config(self.context, provision=True)

            mock_mgr_provision.assert_called_once_with(
                self.context, {"key_derivation_method": "derivation_method"}
            )

    async def test_update_version_to_storage(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            record = StorageRecord(
                "acapy_version",
                "v0.7.2",
            )
            await storage.add_record(record)
            await test_module.add_or_update_version_to_storage(session)
            records = await storage.find_all_records(RECORD_TYPE_ACAPY_VERSION)
            assert len(records) == 1
            assert records[0].value == f"v{__version__}"

    async def test_version_record_not_found(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            assert (len(await storage.find_all_records(RECORD_TYPE_ACAPY_VERSION))) == 0
            with self.assertRaises(test_module.ConfigError):
                await test_module.add_or_update_version_to_storage(session)



import unittest
from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....protocols.didcomm_prefix import DIDCommPrefix
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import message_types as test_message_types
from .. import routes as test_module
from ..messages import Hangup, Rotate
from ..tests import MockConnRecord, test_conn_id

test_valid_rotate_request = {
    "to_did": "did:example:newdid",
}


def generate_mock_hangup_message():
    msg = Hangup(_id="test-message-id")
    return msg


def generate_mock_rotate_message():
    msg = Rotate(_id="test-message-id", **test_valid_rotate_request)
    return msg


class TestDIDRotateRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    @mock.patch.object(
        test_module.ConnRecord,
        "retrieve_by_id",
        return_value=MockConnRecord(test_conn_id, True),
    )
    @mock.patch.object(
        test_module,
        "DIDRotateManager",
        autospec=True,
        return_value=mock.MagicMock(
            rotate_my_did=mock.CoroutineMock(return_value=generate_mock_rotate_message()),
            ensure_supported_did=mock.CoroutineMock(),
        ),
    )
    async def test_rotate(self, *_):
        self.request.match_info = {"conn_id": test_conn_id}
        self.request.json = mock.CoroutineMock(return_value=test_valid_rotate_request)

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            await test_module.rotate(self.request)

            mock_response.assert_called_once_with(
                {
                    "@id": "test-message-id",
                    "@type": DIDCommPrefix.NEW.value + "/" + test_message_types.ROTATE,
                    **test_valid_rotate_request,
                }
            )

    @mock.patch.object(
        test_module.ConnRecord,
        "retrieve_by_id",
        return_value=MockConnRecord(test_conn_id, True),
    )
    @mock.patch.object(
        test_module,
        "DIDRotateManager",
        autospec=True,
        return_value=mock.MagicMock(
            hangup=mock.CoroutineMock(return_value=generate_mock_hangup_message())
        ),
    )
    async def test_hangup(self, *_):
        self.request.match_info = {"conn_id": test_conn_id}
        self.request.json = mock.CoroutineMock(return_value=test_valid_rotate_request)

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            await test_module.hangup(self.request)

            mock_response.assert_called_once_with(
                {
                    "@id": "test-message-id",
                    "@type": DIDCommPrefix.NEW.value + "/" + test_message_types.HANGUP,
                }
            )

    @mock.patch.object(
        test_module,
        "DIDRotateManager",
        autospec=True,
        return_value=mock.MagicMock(
            ensure_supported_did=mock.CoroutineMock(),
        ),
    )
    async def test_rotate_conn_not_found(self, *_):
        self.request.match_info = {"conn_id": test_conn_id}
        self.request.json = mock.CoroutineMock(return_value=test_valid_rotate_request)

        with mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            mock.CoroutineMock(side_effect=StorageNotFoundError()),
        ):
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.rotate(self.request)

    async def test_rotate_did_validation_errors(self):
        self.request.match_info = {"conn_id": test_conn_id}
        self.request.json = mock.CoroutineMock(return_value=test_valid_rotate_request)

        for error_class in [
            test_module.UnsupportedDIDMethodError,
            test_module.UnresolvableDIDError,
            test_module.UnresolvableDIDCommServicesError,
        ]:
            with mock.patch.object(
                test_module,
                "DIDRotateManager",
                autospec=True,
                return_value=mock.MagicMock(
                    ensure_supported_did=mock.CoroutineMock(
                        side_effect=error_class("test error")
                    ),
                ),
            ):
                with self.assertRaises(test_module.web.HTTPBadRequest):
                    await test_module.rotate(self.request)


if __name__ == "__main__":
    unittest.main()



"""Classes for managing a collection of decorators."""

from collections import OrderedDict
from typing import Mapping, Optional, Sequence, Type

from marshmallow import Schema
from marshmallow.fields import Field

from ...core.error import BaseError
from ..models.base import BaseModel

DECORATOR_PREFIX = "~"


class DecoratorError(BaseError):
    """Base error for decorator issues."""


class BaseDecoratorSet(OrderedDict):
    """Collection of decorators."""

    def __init__(self, models: Optional[dict] = None):
        """Initialize a decorator set."""
        self._fields = OrderedDict()
        self._models: Mapping[str, Type[BaseModel]] = models.copy() if models else {}
        self._prefix = DECORATOR_PREFIX

    def copy(self) -> "BaseDecoratorSet":
        """Return a copy of the decorator set."""
        result = super().copy()
        result._fields = OrderedDict(
            (name, field.copy()) for (name, field) in self._fields.items()
        )
        result._models = self._models.copy()
        result._prefix = self._prefix
        return result

    def __eq__(self, other: "BaseDecoratorSet") -> bool:
        """Comparison between base decorator sets."""
        return (
            self._fields == other._fields
            and self._models == other._models
            and self._prefix == other._prefix
            and super().__eq__(other)
        )

    def _init_field(self) -> "BaseDecoratorSet":
        """Create a nested decorator set for a named field."""
        return self.__class__(self._models)

    def field(self, name: str) -> "BaseDecoratorSet":
        """Access a named decorated field."""
        if name not in self._fields:
            self._fields[name] = self._init_field()
        return self._fields[name]

    def has_field(self, name: str) -> bool:
        """Check for the existence of a named decorator field."""
        return bool(self._fields.get(name))

    def remove_field(self, name: str):
        """Remove a named decorated field."""
        if name in self._fields:
            del self._fields[name]

    @property
    def fields(self) -> OrderedDict:
        """Acessor for the set of currently defined fields."""
        return OrderedDict(
            (name, field) for (name, field) in self._fields.items() if field
        )

    @property
    def models(self) -> dict:
        """Accessor for the models dictionary."""
        return self._models.copy()

    @property
    def prefix(self) -> str:
        """Accessor for the decorator prefix."""
        return self._prefix

    def add_model(self, key: str, model: Type[BaseModel]):
        """Add a registered decorator model."""
        self._models[key] = model

    def remove_model(self, key: str):
        """Remove a registered decorator model."""
        del self._models[key]

    def __setitem__(self, key, value):
        """Add a decorator."""
        self.load_decorator(key, value)

    def load_decorator(self, key: str, value, serialized=False):
        """Convert a decorator value to its loaded representation."""
        if key in self._models and isinstance(value, (dict, OrderedDict)):
            if serialized:
                value = self._models[key].deserialize(value)
            else:
                value = self._models[key](**value)
        if value is not None:
            super().__setitem__(key, value)
        elif key in self:
            del self[key]

    def extract_decorators(
        self,
        message: Mapping,
        schema: Type[Schema] = None,
        serialized: bool = True,
        skip_attrs: Sequence[str] = None,
    ) -> OrderedDict:
        """Extract decorators and return the remaining properties."""
        remain = OrderedDict()
        skip_attrs = set(skip_attrs) if skip_attrs else set()
        if schema:
            for field_name, field_def in schema._declared_fields.items():
                if isinstance(field_def, Field) and field_def.data_key:
                    skip_attrs.add(field_def.data_key)
        if message:
            pfx_len = len(self._prefix)
            for key, value in message.items():
                if key in skip_attrs:
                    pass
                elif key.startswith(self._prefix):
                    key = key[pfx_len:]
                    self.load_decorator(key, value, serialized)
                    continue
                elif self._prefix in key:
                    field, key = key.split(self._prefix, 1)
                    self.field(field).load_decorator(key, value, serialized)
                    continue
                remain[key] = value
        return remain

    def to_dict(self, prefix: Optional[str] = None) -> OrderedDict:
        """Convert to a dictionary (serialize).

        Raises:
            BaseModelError: on decorator validation errors

        """
        if prefix is None:
            prefix = self._prefix
        result = OrderedDict()
        for k in self:
            value = self[k]
            if isinstance(value, BaseModel):
                value = value.serialize()
            result[prefix + k] = value
        for k in self._fields:
            result.update(self._fields[k].to_dict(k + prefix))
        return result

    def __repr__(self) -> str:
        """Create a string representation of the decorator set."""
        items = ("{}: {}".format(self._prefix + k, repr(self[k])) for k in self)
        return "<{}{{{}}}>".format(self.__class__.__name__, ", ".join(items))



"""DataIntegrity class."""

from datetime import datetime

from ...core.error import BaseError
from ...core.profile import ProfileSession
from ...resolver.base import DIDNotFound
from .cryptosuites import EddsaJcs2022
from .errors import PROBLEM_DETAILS
from .models.options import DataIntegrityProofOptions
from .models.proof import DataIntegrityProof
from .models.verification_response import (
    DataIntegrityVerificationResponse,
    DataIntegrityVerificationResult,
    ProblemDetails,
)

CRYPTOSUITES = {
    "eddsa-jcs-2022": EddsaJcs2022,
}

PROOF_TYPES = ["DataIntegrityProof"]

PROOF_PURPOSES = [
    "authentication",
    "assertionMethod",
]


class DataIntegrityManagerError(BaseError):
    """Generic DataIntegrityManager Error."""


class DataIntegrityManager:
    """Class for managing data integrity proofs."""

    def __init__(self, session: ProfileSession):
        """Initialize the DataIntegrityManager."""
        self.session = session

    async def add_proof(self, document: dict, options: DataIntegrityProofOptions):
        """Data integrity add proof algorithm.

        https://www.w3.org/TR/vc-data-integrity/#add-proof.
        """
        self.validate_proof_options(options)
        suite = self.select_suite(options)

        # Capture existing proofs if any
        all_proofs = document.pop("proof", [])
        if not isinstance(all_proofs, list) and not isinstance(all_proofs, dict):
            raise DataIntegrityManagerError("Expected proof to be a list or an object.")

        all_proofs = [all_proofs] if isinstance(all_proofs, dict) else all_proofs

        # Create secured document and create new proof
        secured_document = document.copy()
        secured_document["proof"] = all_proofs
        proof = await suite.create_proof(document, options)
        secured_document["proof"].append(proof.serialize())
        return secured_document

    async def verify_proof(self, secured_document: dict):
        """Verify a proof attached to a secured document.

        https://www.w3.org/TR/vc-data-integrity/#verify-proof.
        """
        unsecured_document = secured_document.copy()
        all_proofs = unsecured_document.pop("proof")
        all_proofs = all_proofs if isinstance(all_proofs, list) else [all_proofs]
        verification_results = []
        for proof in all_proofs:
            proof_options = proof.copy()
            proof_options.pop("proofValue")
            proof_options = DataIntegrityProofOptions.deserialize(proof_options)
            try:
                self.validate_proof_options(proof_options)
                suite = self.select_suite(proof_options)
                input_document = unsecured_document.copy()
                input_document["proof"] = proof
                verification_result = await suite.verify_proof(input_document)

            except (DataIntegrityManagerError, DIDNotFound) as err:
                problem_detail = ProblemDetails.deserialize(
                    PROBLEM_DETAILS["PROOF_VERIFICATION_ERROR"]
                )
                problem_detail.detail = str(err)
                verification_result = DataIntegrityVerificationResult(
                    verified=False,
                    proof=DataIntegrityProof.deserialize(proof),
                    problem_details=[problem_detail],
                )
            verification_results.append(verification_result)
        return DataIntegrityVerificationResponse(
            verified=(
                True if all(result.verified for result in verification_results) else False
            ),
            verified_document=unsecured_document,
            results=verification_results,
        )

    def select_suite(self, options: DataIntegrityProofOptions):
        """Instanciate a cryptographic suite.

        https://www.w3.org/TR/vc-data-integrity/#cryptographic-suites.
        """
        if options.type == "DataIntegrityProof":
            suite = CRYPTOSUITES[options.cryptosuite](session=self.session)

        elif options.type in PROOF_TYPES:
            # TODO add support for Ed25519Signature2020
            pass

        else:
            raise DataIntegrityManagerError(f"Unsupported proof type {options.type}")
        return suite

    def validate_proof_options(self, proof_options: DataIntegrityProofOptions):
        """Generic proof assertions for a data integrity proof options."""
        if proof_options.created:
            try:
                datetime.fromisoformat(proof_options.created)
            except ValueError:
                raise DataIntegrityManagerError(
                    f"Invalid proof creation datetime format {proof_options.created}"
                )
        if proof_options.expires:
            try:
                datetime.fromisoformat(proof_options.expires)
            except ValueError:
                raise DataIntegrityManagerError(
                    f"Invalid proof expiration datetime format {proof_options.expires}"
                )
        if proof_options.type not in PROOF_TYPES:
            raise DataIntegrityManagerError(
                f"Unsupported proof type {proof_options.type}"
            )
        if proof_options.type == "DataIntegrityProof":
            if not proof_options.cryptosuite:
                raise DataIntegrityManagerError(
                    "DataIntegrityProof must specify a cryptosuite."
                )
            if proof_options.cryptosuite not in CRYPTOSUITES:
                raise DataIntegrityManagerError(
                    f"Unsupported cryptosuite {proof_options.cryptosuite}"
                )
        if proof_options.proof_purpose not in PROOF_PURPOSES:
            raise DataIntegrityManagerError(
                f"Unsupported proof purpose {proof_options.proof_purpose}"
            )



from unittest import TestCase

from ...wallet.key_type import BLS12381G1
from ...wallet.util import b58_to_bytes
from ..did_key import DID_KEY_RESOLVERS, DIDKey
from .test_dids import (
    DID_BLS12381G1_z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA,
)

TEST_BLS12381G1_BASE58_KEY = (
    "6FywSzB5BPd7xehCo1G4nYHAoZPMMP3gd4PLnvgA6SsTsogtz8K7RDznqLpFPLZXAE"
)
TEST_BLS12381G1_FINGERPRINT = (
    "z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA"
)
TEST_BLS12381G1_DID = f"did:key:{TEST_BLS12381G1_FINGERPRINT}"
TEST_BLS12381G1_KEY_ID = f"{TEST_BLS12381G1_DID}#{TEST_BLS12381G1_FINGERPRINT}"
TEST_BLS12381G1_PREFIX_BYTES = b"".join(
    [b"\xea\x01", b58_to_bytes(TEST_BLS12381G1_BASE58_KEY)]
)


class TestDIDKey(TestCase):
    def test_bls12381g1_from_public_key(self):
        key_bytes = b58_to_bytes(TEST_BLS12381G1_BASE58_KEY)
        did_key = DIDKey.from_public_key(key_bytes, BLS12381G1)

        assert did_key.did == TEST_BLS12381G1_DID

    def test_bls12381g1_from_public_key_b58(self):
        did_key = DIDKey.from_public_key_b58(TEST_BLS12381G1_BASE58_KEY, BLS12381G1)

        assert did_key.did == TEST_BLS12381G1_DID

    def test_bls12381g1_from_fingerprint(self):
        did_key = DIDKey.from_fingerprint(TEST_BLS12381G1_FINGERPRINT)

        assert did_key.did == TEST_BLS12381G1_DID
        assert did_key.public_key_b58 == TEST_BLS12381G1_BASE58_KEY

    def test_bls12381g1_from_did(self):
        did_key = DIDKey.from_did(TEST_BLS12381G1_DID)

        assert did_key.public_key_b58 == TEST_BLS12381G1_BASE58_KEY

    def test_bls12381g1_properties(self):
        did_key = DIDKey.from_did(TEST_BLS12381G1_DID)

        assert did_key.fingerprint == TEST_BLS12381G1_FINGERPRINT
        assert did_key.did == TEST_BLS12381G1_DID
        assert did_key.public_key_b58 == TEST_BLS12381G1_BASE58_KEY
        assert did_key.public_key == b58_to_bytes(TEST_BLS12381G1_BASE58_KEY)
        assert did_key.key_type == BLS12381G1
        assert did_key.key_id == TEST_BLS12381G1_KEY_ID
        assert did_key.prefixed_public_key == TEST_BLS12381G1_PREFIX_BYTES

    def test_bls12381g1_diddoc(self):
        did_key = DIDKey.from_did(TEST_BLS12381G1_DID)

        resolver = DID_KEY_RESOLVERS[BLS12381G1]

        assert resolver(did_key) == did_key.did_doc

    def test_bls12381g1_resolver(self):
        did_key = DIDKey.from_did(TEST_BLS12381G1_DID)
        resolver = DID_KEY_RESOLVERS[BLS12381G1]
        did_doc = resolver(did_key)

        assert (
            did_doc
            == DID_BLS12381G1_z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA
        )



"""Test mediate grant message."""

from unittest import TestCase

from ...message_types import MEDIATE_GRANT
from ..mediate_grant import MediationGrant, MediationGrantSchema
from . import MessageTest


class TestMediateGrant(MessageTest, TestCase):
    """Test mediate grant message."""

    TYPE = MEDIATE_GRANT
    CLASS = MediationGrant
    SCHEMA = MediationGrantSchema
    VALUES = {
        "endpoint": "http://localhost:3000",
        "routing_keys": ["did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"],
    }



"""This class represents schema information for anoncreds."""

from typing import Optional


class AnonCredsSchemaInfo:
    """Represents the schema information for anonymous credentials.

    Attributes:
        issuer_id (str): The identifier of the issuer.
        name (Optional[str]): The name of the schema. Defaults to None.
        version (Optional[str]): The version of the schema. Defaults to None.

    Args:
        issuer_id (str): The identifier of the issuer.
        name (Optional[str], optional): The name of the schema. Defaults to None.
        version (Optional[str], optional): The version of the schema. Defaults to None.
    """

    def __init__(
        self, issuer_id: str, name: Optional[str] = None, version: Optional[str] = None
    ):
        """Initialize the schema information."""
        self.issuer_id = issuer_id
        self.name = name
        self.version = version



"""Message type identifiers for Connections."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "527849ec3aa2a8fd47a7bb6c57f918ff8bcb5e8c/features/0095-basic-message"
)

# Message types
BASIC_MESSAGE = "basicmessage/1.0/message"

PROTOCOL_PACKAGE = "acapy_agent.protocols.basicmessage.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {BASIC_MESSAGE: f"{PROTOCOL_PACKAGE}.messages.basicmessage.BasicMessage"}
)



"""Transport-related error classes and codes."""

from ..core.error import BaseError


class TransportError(BaseError):
    """Base class for all transport errors."""


class WireFormatError(TransportError):
    """Base class for wire-format errors."""


class WireFormatParseError(WireFormatError):
    """Parse error when unpacking the wire format."""

    error_code = "message_parse_error"


class WireFormatEncodeError(WireFormatError):
    """Encoding error when packing the wire format."""

    error_code = "message_encode_error"


class RecipientKeysError(WireFormatError):
    """Extract recipient keys error."""
